from xline.base_classes import Element
from .gaussian_fields import get_Ex_Ey_Gx_Gy_gauss
from .qgauss import QGauss
from scipy.interpolate import CubicSpline


class SCCoasting(Element):
    """Space charge for a coasting beam."""

    _description = [
        ("number_of_particles", "", "Number of particles in the beam", 0.0),
        ("circumference", "m", "Machine circumference", 1.0),
        ("sigma_x", "m", "Horizontal size of the beam (r.m.s.)", 1.0),
        ("sigma_y", "m", "Vertical size of the beam (r.m.s.)", 1.0),
        ("length", "m", "Integration length of space charge kick", 0.0),
        ("x_co", "m", "Horizontal closed orbit offset", 0.0),
        ("y_co", "m", "Vertical closed orbit offset", 0.0),
    ]
    _extra = [
        ("min_sigma_diff", "m", "Threshold to detect round beam", 1e-8),
        ("enabled", "", "Switch to disable space charge effect", True),
    ]

    def track(self, p):
        if self.enabled:
            charge = p.q0 * p.echarge

            Ex, Ey = get_Ex_Ey_Gx_Gy_gauss(
                p.x - self.x_co,
                p.y - self.y_co,
                self.sigma_x,
                self.sigma_y,
                min_sigma_diff=self.min_sigma_diff,
                skip_Gs=True,
                mathlib=p._m,
            )

            fact_kick = (
                p.chi
                * self.number_of_particles
                / self.circumference
                * (charge * p.charge_ratio)
                * charge
                * (1 - p.beta0 * p.beta0)
                / (p.p0c * p.echarge * p.beta0)
                * self.length
            )

            p.px += fact_kick * Ex
            p.py += fact_kick * Ey


class SCQGaussProfile(Element):
    """Space charge for a bunched beam with generalised
    Gaussian profile.
    """

    _description = [
        ("number_of_particles", "", "Number of particles in the bunch", 0.0),
        ("bunchlength_rms", "m", "Length of the bunch (r.m.s.)", 1.0),
        ("sigma_x", "m", "Horizontal size of the beam (r.m.s.)", 1.0),
        ("sigma_y", "m", "Vertical size of the beam (r.m.s.)", 1.0),
        ("length", "m", "Integration length of space charge kick", 0.0),
        ("x_co", "m", "Horizontal closed orbit offset", 0.0),
        ("y_co", "m", "Vertical closed orbit offset", 0.0),
    ]
    _extra = [
        ("min_sigma_diff", "m", "Threshold to detect round beam", 1e-8),
        ("enabled", "", "Switch to disable space charge effect", True),
        (
            "q_parameter",
            "",
            "q parameter of generalised Gaussian distribution (q=1 for standard Gaussian)",
            1.0,
        ),
    ]

    def track(self, p):
        if self.enabled:
            distr = QGauss(self.q_parameter, mathlib=p._m)
            sigma = p.zeta / p.rvv
            fact_kick = self.number_of_particles * distr.eval(
                sigma, QGauss.sqrt_beta(self.bunchlength_rms)
            )

            charge = p.q0 * p.echarge
            fact_kick *= p.chi * p.charge_ratio * self.length * charge * charge
            fact_kick *= 1 - p.beta0 * p.beta0
            fact_kick /= p.p0c * p.echarge * p.beta0

            Ex, Ey = get_Ex_Ey_Gx_Gy_gauss(
                p.x - self.x_co,
                p.y - self.y_co,
                self.sigma_x,
                self.sigma_y,
                min_sigma_diff=self.min_sigma_diff,
                skip_Gs=True,
                mathlib=p._m,
            )

            p.px += fact_kick * Ex
            p.py += fact_kick * Ey


class SCInterpolatedProfile(Element):
    """Space charge for a bunched beam with discretised profile."""

    _description = [
        ("number_of_particles", "", "Number of particles in the bunch", 0.0),
        (
            "line_density_profile",
            "1/m",
            "Discretised list of density values with integral normalised to 1",
            lambda: [1.0, 1.0],
        ),
        ("dz", "m", "Unit distance in zeta between profile points", 1.0),
        ("z0", "m", "Start zeta position of line density profile", -0.5),
        ("sigma_x", "m", "Horizontal size of the beam (r.m.s.)", 1.0),
        ("sigma_y", "m", "Vertical size of the beam (r.m.s.)", 1.0),
        ("length", "m", "Integration length of space charge kick", 0.0),
        ("x_co", "m", "Horizontal closed orbit offset", 0.0),
        ("y_co", "m", "Vertical closed orbit offset", 0.0),
    ]
    _extra = [
        (
            "method",
            "",
            "Interpolation method; 0 == linear (default), 1 == cubic spline",
            0,
        ),
        ("min_sigma_diff", "m", "Threshold to detect round beam", 1e-8),
        ("enabled", "", "Switch to disable space charge effect", True),
    ]

    def track(self, p):
        if self.enabled:
            n_prof_points = len(self.line_density_profile)
            charge = p.q0 * p.echarge

            Ex, Ey = get_Ex_Ey_Gx_Gy_gauss(
                p.x - self.x_co,
                p.y - self.y_co,
                self.sigma_x,
                self.sigma_y,
                min_sigma_diff=self.min_sigma_diff,
                skip_Gs=True,
                mathlib=p._m,
            )

            fact_kick = (
                p.chi
                * (charge * p.charge_ratio)
                * charge
                * (1 - p.beta0 * p.beta0)
                / (p.p0c * p.echarge * p.beta0)
                * self.length
            )

            absc_values = p._m.linspace(
                self.z0, self.z0 + self.dz * (n_prof_points - 1), n_prof_points
            )

            if self.method == 0:
                ld_factor = p._m.interp(
                    p.zeta, absc_values, self.line_density_profile
                )
            elif self.method == 1:
                cs = CubicSpline(absc_values, self.line_density_profile)
                ld_factor = cs(p.zeta)
            else:
                ld_factor = 1

            fact_kick *= self.number_of_particles * ld_factor
            p.px += fact_kick * Ex
            p.py += fact_kick * Ey



import numpy as np

from .base_classes import Element
from .be_beamfields.beambeam import BeamBeam4D
from .be_beamfields.beambeam import BeamBeam6D
from .be_beamfields.spacecharge import SCCoasting
from .be_beamfields.spacecharge import SCQGaussProfile
from .be_beamfields.spacecharge import SCInterpolatedProfile

_factorial = np.array(
    [
        1,
        1,
        2,
        6,
        24,
        120,
        720,
        5040,
        40320,
        362880,
        3628800,
        39916800,
        479001600,
        6227020800,
        87178291200,
        1307674368000,
        20922789888000,
        355687428096000,
        6402373705728000,
        121645100408832000,
        2432902008176640000,
    ]
)


class Drift(Element):
    """Drift in expanded form"""

    _description = [("length", "m", "Length of the drift", 0)]

    def track(self, p):
        length = self.length
        rpp = p.rpp
        xp = p.px * rpp
        yp = p.py * rpp
        p.x += xp * length
        p.y += yp * length
        p.zeta += length * (p.rvv - (1 + (xp ** 2 + yp ** 2) / 2))
        p.s += length


class DriftExact(Drift):
    """Drift in exact form"""

    _description = [("length", "m", "Length of the drift", 0)]

    def track(self, p):
        sqrt = p._m.sqrt
        length = self.length
        opd = 1 + p.delta
        lpzi = length / sqrt(opd ** 2 - p.px ** 2 - p.py ** 2)
        p.x += p.px * lpzi
        p.y += p.py * lpzi
        p.zeta += p.rvv * length - opd * lpzi
        p.s += length


def _arrayofsize(ar, size):
    ar = np.array(ar)
    if len(ar) == 0:
        return np.zeros(size, dtype=ar.dtype)
    elif len(ar) < size:
        ar = np.hstack([ar, np.zeros(size - len(ar), dtype=ar.dtype)])
    return ar


class Multipole(Element):
    """ Multipole """

    _description = [
        (
            "knl",
            "m^-n",
            "Normalized integrated strength of normal components",
            lambda: [0],
        ),
        (
            "ksl",
            "m^-n",
            "Normalized integrated strength of skew components",
            lambda: [0],
        ),
        (
            "hxl",
            "rad",
            "Rotation angle of the reference trajectory"
            "in the horizzontal plane",
            0,
        ),
        (
            "hyl",
            "rad",
            "Rotation angle of the reference trajectory in the vertical plane",
            0,
        ),
        ("length", "m", "Length of the originating thick multipole", 0),
    ]

    @property
    def order(self):
        return max(len(self.knl), len(self.ksl)) - 1

    def track(self, p):
        order = self.order
        length = self.length
        knl = _arrayofsize(self.knl, order + 1)
        ksl = _arrayofsize(self.ksl, order + 1)
        x = p.x
        y = p.y
        chi = p.chi
        dpx = knl[order]
        dpy = ksl[order]
        for ii in range(order, 0, -1):
            zre = (dpx * x - dpy * y) / ii
            zim = (dpx * y + dpy * x) / ii
            dpx = knl[ii - 1] + zre
            dpy = ksl[ii - 1] + zim
        dpx = -chi * dpx
        dpy = chi * dpy
        # curvature effect kick
        hxl = self.hxl
        hyl = self.hyl
        delta = p.delta
        if hxl != 0 or hyl != 0:
            b1l = chi * knl[0]
            a1l = chi * ksl[0]
            hxlx = hxl * x
            hyly = hyl * y
            if length > 0:
                hxx = hxlx / length
                hyy = hyly / length
            else:  # non physical weak focusing disabled (SixTrack mode)
                hxx = 0
                hyy = 0
            dpx += hxl + hxl * delta - b1l * hxx
            dpy -= hyl + hyl * delta - a1l * hyy
            p.zeta -= chi * (hxlx - hyly)
        p.px += dpx
        p.py += dpy


class RFMultipole(Element):
    """
    H= -l sum   Re[ (kn[n](zeta) + i ks[n](zeta) ) (x+iy)**(n+1)/ n ]

    kn[n](z) = k_n cos(2pi w tau + pn/180*pi)
    ks[n](z) = k_n cos(2pi w tau + pn/180*pi)

    """

    _description = [
        ("voltage", "volt", "Voltage", 0),
        ("frequency", "hertz", "Frequency", 0),
        ("lag", "degree", "Delay in the cavity sin(lag - w tau)", 0),
        ("knl", "", "...", lambda: [0]),
        ("ksl", "", "...", lambda: [0]),
        ("pn", "", "...", lambda: [0]),
        ("ps", "", "...", lambda: [0]),
    ]

    @property
    def order(self):
        return max(len(self.knl), len(self.ksl)) - 1

    def track(self, p):
        sin = p._m.sin
        cos = p._m.cos
        pi = p._m.pi
        order = self.order
        k = 2 * pi * self.frequency / p.clight
        tau = p.zeta / p.rvv / p.beta0
        ktau = k * tau
        deg2rad = pi / 180
        knl = _arrayofsize(self.knl, order + 1)
        ksl = _arrayofsize(self.ksl, order + 1)
        pn = _arrayofsize(self.pn, order + 1) * deg2rad
        ps = _arrayofsize(self.ps, order + 1) * deg2rad
        x = p.x
        y = p.y
        dpx = 0
        dpy = 0
        dptr = 0
        zre = 1
        zim = 0
        for ii in range(order + 1):
            pn_ii = pn[ii] - ktau
            ps_ii = ps[ii] - ktau
            cn = cos(pn_ii)
            sn = sin(pn_ii)
            cs = cos(ps_ii)
            ss = sin(ps_ii)
            # transverse kick order i!
            dpx += cn * knl[ii] * zre - cs * ksl[ii] * zim
            dpy += cs * ksl[ii] * zre + cn * knl[ii] * zim
            # compute z**(i+1)/(i+1)!
            zret = (zre * x - zim * y) / (ii + 1)
            zim = (zim * x + zre * y) / (ii + 1)
            zre = zret
            fnr = knl[ii] * zre
            # fni = knl[ii] * zim
            # fsr = ksl[ii] * zre
            fsi = ksl[ii] * zim
            # energy kick order i+1
            dptr += sn * fnr - ss * fsi

        chi = p.chi
        p.px += -chi * dpx
        p.py += chi * dpy
        dv0 = self.voltage * sin(self.lag * deg2rad - ktau)
        p.add_to_energy(p.charge_ratio * p.q0 * (dv0 - p.p0c * k * dptr))


class Cavity(Element):
    """Radio-frequency cavity"""

    _description = [
        ("voltage", "V", "Integrated energy change", 0),
        ("frequency", "Hz", "Frequency of the cavity", 0),
        ("lag", "degree", "Delay in the cavity sin(lag - w tau)", 0),
    ]

    def track(self, p):
        sin = p._m.sin
        pi = p._m.pi
        k = 2 * pi * self.frequency / p.clight
        tau = p.zeta / p.rvv / p.beta0
        phase = self.lag * pi / 180 - k * tau
        p.add_to_energy(p.charge_ratio * p.q0 * self.voltage * sin(phase))


class SawtoothCavity(Element):
    """Radio-frequency cavity"""

    _description = [
        ("voltage", "V", "Integrated energy change", 0),
        ("frequency", "Hz", "Equivalent Frequency of the cavity", 0),
        ("lag", "degree", "Delay in the cavity `lag - w tau`", 0),
    ]

    def track(self, p):
        pi = p._m.pi
        k = 2 * pi * self.frequency / p.clight
        tau = p.zeta / p.rvv / p.beta0
        phase = self.lag * pi / 180 - k * tau
        phase = (phase + pi) % (2 * pi) - pi
        p.add_to_energy(p.charge_ratio * p.q0 * self.voltage * phase)


class XYShift(Element):
    """shift of the reference"""

    _description = [
        ("dx", "m", "Horizontal shift", 0),
        ("dy", "m", "Vertical shift", 0),
    ]

    def track(self, p):
        p.x -= self.dx
        p.y -= self.dy




class Elens(Element):
    """Hollow Electron Lens"""

    _description = [("voltage", "V", "Voltage of the electron lens", 0),
                    ("current", "A", "Current of the e-beam", 0),
                    ("inner_radius", "m", "Inner radius of the hollow e-beam", 0),
                    ("outer_radius", "m", "Outer radius of the hollow e-beam", 0),
                    ("ebeam_center_x", "m", "Center of the e-beam in x", 0),
                    ("ebeam_center_y", "m", "Center of the e-beam in y", 0),
                    ("elens_length", "m", "Length of the hollow electron lens", 0)
                    ]

    def track(self, p):

        # vacuum permittivity
        epsilon0 = p.epsilon0
        pi       = p._m.pi             # pi
        clight   = p.clight            # speed of light
        e_mass   = p.emass             # electron mass

        # get the transverse amplitude
        # TO DO: needs to be modified for off-centererd e-beam
        r = np.sqrt(p.x**2 + p.y**2)

        # magnetic rigidity

        if type(p.pc) is float:
            Brho = p.pc/(p.q0*p.clight)
        else:
            Brho = p.pc[0]/(p.q0*p.clight)


        # Electron properties
        Ekin_e = self.voltage                         # kinetic energy
        Etot_e = Ekin_e + e_mass                      # total energy
        p_e    = np.sqrt(Etot_e**2 - e_mass**2)       # electron momentum
        beta_e = p_e/Etot_e                           # relativ. beta

        # relativistic beta  of protons
        beta_p = p.rvv*p.beta0

        # abbreviate for better readability
        r1 = self.inner_radius
        r2 = self.outer_radius
        I  = self.current

        # geometric factor frr
        frr = ((r**2 - r1**2)/(r2**2 - r1**2))  # uniform distribution

        try:
            frr = [max(0,iitem) for iitem in frr]
            frr = [min(1,iitem) for iitem in frr]
            frr = np.array(frr, dtype = float)

        except TypeError:
            frr = max(0,frr)
            frr = min(1,frr)
            frr = np.array([frr], dtype=float)


        #
        #
        # if len(frr)>0:
        #     frr[frr<0] = 0
        #     frr[frr>1] = 1

        # calculate the kick at r2 (maximum kick)
        theta_max = ((1/(4*pi*epsilon0))*(2*self.elens_length*I)*
                    (1+beta_e*beta_p)*(1/(r2*Brho*beta_e*beta_p*clight**2)))

        # calculate the kick of the particles
        # the (-1) stems from the attractive force of the E-field
        theta = (-1)*theta_max*r2*p.rpp*p.chi

        print("type frr", type(frr))
        print("type r", type(r))

        theta = theta*np.divide(frr, r, out=np.zeros_like(frr), where=r!=0)

        # convert px and py to x' and y'
        xp   = p.px * p.rpp
        yp   = p.py * p.rpp

        # update xp and yp with the HEL kick
        # use np.divide to not crash when r=0
        xp = xp + p.x*np.divide(theta, r, out=np.zeros_like(theta), where=r!=0)
        yp = yp + p.y*np.divide(theta, r, out=np.zeros_like(theta), where=r!=0)

        # update px and py.
        p.px = xp/p.rpp
        p.py = yp/p.rpp



class SRotation(Element):
    """anti-clockwise rotation of the reference frame"""

    _description = [("angle", "", "Rotation angle", 0)]

    def track(self, p):
        deg2rag = p._m.pi / 180
        cz = p._m.cos(self.angle * deg2rag)
        sz = p._m.sin(self.angle * deg2rag)
        xn = cz * p.x + sz * p.y
        yn = -sz * p.x + cz * p.y
        p.x = xn
        p.y = yn
        xn = cz * p.px + sz * p.py
        yn = -sz * p.px + cz * p.py
        p.px = xn
        p.py = yn


class LimitRect(Element):
    _description = [
        ("min_x", "m", "Minimum horizontal aperture", -1.0),
        ("max_x", "m", "Maximum horizontal aperture", 1.0),
        ("min_y", "m", "Minimum vertical aperture", -1.0),
        ("max_y", "m", "Minimum vertical aperture", 1.0),
    ]

    def track(self, particle):

        x = particle.x
        y = particle.y

        if not hasattr(particle.state, "__iter__"):
            particle.state = int(
                x >= self.min_x
                and x <= self.max_x
                and y >= self.min_y
                and y <= self.max_y
            )
            if particle.state != 1:
                return "Particle lost"
        else:
            particle.state = np.int_(
                (x >= self.min_x)
                & (x <= self.max_x)
                & (y >= self.min_y)
                & (y <= self.max_y)
            )
            particle.remove_lost_particles()
            if len(particle.state) == 0:
                return "All particles lost"


class LimitEllipse(Element):
    _description = [
        ("a", "m", "Horizontal semiaxis", 1.0),
        ("b", "m", "Vertical semiaxis", 1.0),
    ]

    def track(self, particle):

        x = particle.x
        y = particle.y

        if not hasattr(particle.state, "__iter__"):
            particle.state = int(
                x * x / (self.a * self.a) + y * y / (self.b * self.b) <= 1.0
            )
            if particle.state != 1:
                return "Particle lost"
        else:
            particle.state = np.int_(
                x * x / (self.a * self.a) + y * y / (self.b * self.b) <= 1.0
            )
            particle.remove_lost_particles()
            if len(particle.state) == 0:
                return "All particles lost"


class LimitRectEllipse(Element):
    _description = [
        ("max_x", "m", "Maximum horizontal aperture", 1.0),
        ("max_y", "m", "Maximum vertical aperture", 1.0),
        ("a", "m", "Horizontal semiaxis", 1.0),
        ("b", "m", "Vertical semiaxis", 1.0),
    ]

    def track(self, particle):

        x = particle.x
        y = particle.y

        if not hasattr(particle.state, "__iter__"):
            particle.state = int(
                x >= -self.max_x
                and x <= self.max_x
                and y >= -self.max_y
                and y <= self.max_y
                and x * x / (self.a * self.a) + y * y / (self.b * self.b) <= 1.0
            )
            if particle.state != 1:
                return "Particle lost"
        else:
            particle.state = np.int_(
                (x >= -self.max_x)
                & (x <= self.max_x)
                & (y >= -self.max_y)
                & (y <= self.max_y)
                & (x * x / (self.a * self.a) + y * y / (self.b * self.b) <= 1.0)
            )
            particle.remove_lost_particles()
            if len(particle.state) == 0:
                return "All particles lost"


class BeamMonitor(Element):
    _description = [
        ("num_stores", "", "...", 0),
        ("start", "", "...", 0),
        ("skip", "", "...", 1),
        ("max_particle_id", "", "", 0),
        ("min_particle_id", "", "", 0),
        ("is_rolling", "", "", False),
        ("is_turn_ordered", "", "", True),
        ("data", "", "...", lambda: []),
    ]

    def offset(self, particle):
        _offset = -1
        nn = (
            self.max_particle_id >= self.min_particle_id
            and (self.max_particle_id - self.min_particle_id + 1)
            or -1
        )
        assert self.is_turn_ordered

        if (
            particle.turn >= self.start
            and nn > 0
            and particle.particle_id >= self.min_particle_id
            and particle.particle_id <= self.max_particle_id
        ):
            turns_since_start = particle.turns - self.start
            store_index = turns_since_start // self.skip
            if store_index < self.num_stores:
                pass
            elif self.is_rolling:
                store_index = store_index % self.num_stores
            else:
                store_index = -1

            if store_index >= 0:
                _offset = store_index * nn + particle.particle_id

        return _offset

    def track(self, p):
        self.data.append(p.copy)


class DipoleEdge(Element):
    _description = [
        ("h", "1/m", "Curvature", 0),
        ("e1", "rad", "Face angle", 0),
        ("hgap", "m", "Equivalent gap", 0),
        ("fint", "", "Fringe integral", 0),
    ]

    def track(self, p):
        tan = p._m.tan
        sin = p._m.sin
        cos = p._m.cos
        corr = 2 * self.h * self.hgap * self.fint
        r21 = self.h * tan(self.e1)
        r43 = -self.h * tan(
            self.e1 - corr / cos(self.e1) * (1 + sin(self.e1) ** 2)
        )
        p.px += r21 * p.x
        p.py += r43 * p.y


__all__ = [
    "BeamBeam4D",
    "BeamBeam6D",
    "BeamMonitor",
    "Cavity",
    "DipoleEdge",
    "Drift",
    "DriftExact",
    "Element",
    "Elens",
    "LimitEllipse",
    "LimitRect",
    "Multipole",
    "RFMultipole",
    "SCCoasting",
    "SCInterpolatedProfile",
    "SCQGaussProfile",
    "SRotation",
    "XYShift",
]



from math import factorial

import numpy as np

from . import elements

clight = 299792458
pi = np.pi


def bn_mad(bn_mad, n, sign):
    return sign * bn_mad * factorial(n - 1)


def bn_rel(bn16, bn3, r0, d0, sign):
    out = []
    for nn, (a, b) in enumerate(zip(bn16, bn3)):
        n = nn + 1
        sixval = d0 * a * b * r0 ** (1 - n) * 10 ** (3 * n - 6)
        out.append(bn_mad(sixval, n, sign))
    return out


def _expand_struct(sixinput, convert=elements):
    elems = []
    count = {}
    icount = 0
    iconv = []
    names = []
    rest = []
    Drift = convert.Drift
    Multipole = convert.Multipole
    Cavity = convert.Cavity
    XYShift = convert.XYShift
    SRotation = convert.SRotation
    # Line = convert.Line
    BeamBeam4D = convert.BeamBeam4D
    BeamBeam6D = convert.BeamBeam6D
    RFMultipole = convert.RFMultipole
    exclude = False
    # add special elenents
    if "CAV" in sixinput.iter_struct():
        sixinput.single["CAV"] = [
            12 * sixinput.ition,
            sixinput.u0,
            sixinput.harm,
            0,
        ]
    for nnn in sixinput.iter_struct():
        exclude = False
        ccc = count.setdefault(nnn, 0)
        if len(sixinput.single[nnn]) == 7:
            etype, d1, d2, d3, _, _, _ = sixinput.single[nnn]
        else:
            etype, d1, d2, d3, = sixinput.single[nnn]
        elem = None
        if nnn in sixinput.align:
            dx, dy, tilt = sixinput.align[nnn][ccc]
            tilt = tilt * 180e-3 / pi
            dx *= 1e-3
            dy *= 1e-3
            hasshift = abs(dx) + abs(dy) > 0
            hastilt = abs(tilt) > 0
            if hasshift:
                names.append(nnn + "_preshift")
                elems.append(XYShift(dx=dx, dy=dy))
                icount += 1
            if hastilt:
                names.append(nnn + "_pretilt")
                elems.append(SRotation(angle=tilt))
                icount += 1
        if etype in [0, 25]:
            elem = Drift(length=d3)
            if d3 > 0:
                exclude = True
        elif abs(etype) in [1, 2, 3, 4, 5, 7, 8, 9, 10]:
            bn_six = d1
            nn = abs(etype)
            sign = -etype / nn
            madval = bn_mad(bn_six, nn, sign)
            knl = [0] * (nn - 1) + [madval]
            ksl = [0] * nn
            if sign == 1:
                knl, ksl = ksl, knl
            elem = Multipole(knl=knl, ksl=ksl, hxl=0, hyl=0, length=0)
        elif etype == 11:
            knl, ksl = sixinput.get_knl(nnn, ccc)
            hxl = 0
            hyl = 0
            length = 0
            # beaware of the case of thick bend
            # see beambeam example where mbw has the length
            if d3 == -1:
                hxl = -d1
                length = d2
                knl[0] = hxl
            elif d3 == -2:
                hyl = -d1  # strange sign!!!
                length = d2
                ksl[0] = hyl
            elem = Multipole(knl=knl, ksl=ksl, hxl=hxl, hyl=hyl, length=length)
        elif etype == 12:
            e0=sixinput.initialconditions[-1]
            p0c=np.sqrt(e0**2-sixinput.pma**2)
            beta0=p0c/e0
            v = d1 * 1e6
            freq = d2 * clight * beta0 / sixinput.tlen
            #print(v,freq)
            elem = Cavity(voltage=v, frequency=freq, lag=180 - d3)
        elif etype == 20:
            thisbb = sixinput.bbelements[nnn]
            if hasattr(thisbb, "sigma_x"):
                elem = BeamBeam4D(**thisbb._asdict())
            elif hasattr(thisbb, "phi"):
                elem = BeamBeam6D(**thisbb._asdict())
            else:
                raise ValueError("What?!")
        elif etype == 23:
            #print(nnn, sixinput.single[nnn])
            voltage_V = d1 *1e6
            freq_Hz = d2 * 1e6
            phase_rad = d3
            p0c_eV = sixinput.initialconditions[12]*1e6
            elem = RFMultipole(
                frequency = freq_Hz,
                knl= [voltage_V / p0c_eV],
                pn=[90.],
                )
        elif etype == -23:
            voltage_V = d1 *1e6
            freq_Hz = d2 * 1e6
            phase_rad = d3
            p0c_eV = sixinput.initialconditions[12]*1e6
            print(nnn, sixinput.single[nnn])
            print(f'p0c_eV: {p0c_eV}')
            elem = RFMultipole(
                frequency = freq_Hz,
                ksl= [-voltage_V / p0c_eV],
                ps=[90.],
                )
        else:
            rest.append([nnn] + sixinput.single[nnn])
        if elem is not None:
            elems.append(elem)
            names.append(nnn)
        if nnn in sixinput.align:
            if hastilt:
                names.append(nnn + "_posttilt")
                elems.append(SRotation(angle=-tilt))
                icount += 1
            if hasshift:
                names.append(nnn + "_postshift")
                elems.append(XYShift(dx=-dx, dy=-dy))
                icount += 1
        if elem is not None:
            if not exclude:
                iconv.append(icount)
            icount += 1
        count[nnn] = ccc + 1
    # newelems = [dict(i._asdict()) for i in elems]
    types = [i.__class__.__name__ for i in elems]
    return list(zip(names, types, elems)), rest, iconv



import numpy as np


def mysign(u):
    return 2 * np.float_(u >= 0) - 1.0


class Sigmas(object):
    def __init__(
        self,
        Sig_11_0,
        Sig_12_0,
        Sig_13_0,
        Sig_14_0,
        Sig_22_0,
        Sig_23_0,
        Sig_24_0,
        Sig_33_0,
        Sig_34_0,
        Sig_44_0,
    ):

        self.Sig_11_0 = Sig_11_0
        self.Sig_12_0 = Sig_12_0
        self.Sig_13_0 = Sig_13_0
        self.Sig_14_0 = Sig_14_0
        self.Sig_22_0 = Sig_22_0
        self.Sig_23_0 = Sig_23_0
        self.Sig_24_0 = Sig_24_0
        self.Sig_33_0 = Sig_33_0
        self.Sig_34_0 = Sig_34_0
        self.Sig_44_0 = Sig_44_0

    def tobuffer(self):
        buf = [
            self.Sig_11_0,
            self.Sig_12_0,
            self.Sig_13_0,
            self.Sig_14_0,
            self.Sig_22_0,
            self.Sig_23_0,
            self.Sig_24_0,
            self.Sig_33_0,
            self.Sig_34_0,
            self.Sig_44_0,
        ]
        return np.array(buf, dtype=np.float64)


def boost_sigmas(Sigma_0, cphi):
    Sigma_0_boosted = Sigmas(
        Sigma_0.Sig_11_0,
        Sigma_0.Sig_12_0 / cphi,
        Sigma_0.Sig_13_0,
        Sigma_0.Sig_14_0 / cphi,
        Sigma_0.Sig_22_0 / cphi / cphi,
        Sigma_0.Sig_23_0 / cphi,
        Sigma_0.Sig_24_0 / cphi / cphi,
        Sigma_0.Sig_33_0,
        Sigma_0.Sig_34_0 / cphi,
        Sigma_0.Sig_44_0 / cphi / cphi,
    )
    return Sigma_0_boosted


def _propagate_Sigma_matrix(
    Sigmas_at_0, S, threshold_singular=1e-16, handle_singularities=True
):

    Sig_11_0 = Sigmas_at_0.Sig_11_0
    Sig_12_0 = Sigmas_at_0.Sig_12_0
    Sig_13_0 = Sigmas_at_0.Sig_13_0
    Sig_14_0 = Sigmas_at_0.Sig_14_0
    Sig_22_0 = Sigmas_at_0.Sig_22_0
    Sig_23_0 = Sigmas_at_0.Sig_23_0
    Sig_24_0 = Sigmas_at_0.Sig_24_0
    Sig_33_0 = Sigmas_at_0.Sig_33_0
    Sig_34_0 = Sigmas_at_0.Sig_34_0
    Sig_44_0 = Sigmas_at_0.Sig_44_0

    # ~ Sig_11 = Sig_11_0 + 2.*Sig_12_0*S+Sig_22_0*S*S
    # ~ Sig_33 = Sig_33_0 + 2.*Sig_34_0*S+Sig_44_0*S*S
    # ~ Sig_13 = Sig_13_0 + (Sig_14_0+Sig_23_0)*S+Sig_24_0*S*S

    (
        Sig_11,
        Sig_12,
        Sig_13,
        Sig_14,
        Sig_22,
        Sig_23,
        Sig_24,
        Sig_33,
        Sig_34,
        Sig_44,
    ) = propagate_full_Sigma_matrix_in_drift(
        Sig_11_0,
        Sig_12_0,
        Sig_13_0,
        Sig_14_0,
        Sig_22_0,
        Sig_23_0,
        Sig_24_0,
        Sig_33_0,
        Sig_34_0,
        Sig_44_0,
        S,
    )

    R = Sig_11 - Sig_33
    W = Sig_11 + Sig_33
    T = R * R + 4 * Sig_13 * Sig_13

    # evaluate derivatives
    dS_R = 2.0 * (Sig_12_0 - Sig_34_0) + 2 * S * (Sig_22_0 - Sig_44_0)
    dS_W = 2.0 * (Sig_12_0 + Sig_34_0) + 2 * S * (Sig_22_0 + Sig_44_0)
    dS_Sig_13 = Sig_14_0 + Sig_23_0 + 2 * Sig_24_0 * S
    dS_T = 2 * R * dS_R + 8.0 * Sig_13 * dS_Sig_13

    signR = mysign(R)

    if T < threshold_singular and handle_singularities:
        a = Sig_12 - Sig_34
        b = Sig_22 - Sig_44
        c = Sig_14 + Sig_23
        d = Sig_24

        sqrt_a2_c2 = np.sqrt(a * a + c * c)

        if sqrt_a2_c2 * sqrt_a2_c2 * sqrt_a2_c2 < threshold_singular:
            # equivalent to: if np.abs(c)<threshold_singular and
            #                   np.abs(a)<threshold_singular:

            if np.abs(d) > threshold_singular:
                cos2theta = np.abs(b) / np.sqrt(b * b + 4 * d * d)
            else:
                cos2theta = 1.0  # Decoupled beam

            costheta = np.sqrt(0.5 * (1.0 + cos2theta))
            sintheta = mysign(b) * mysign(d) * np.sqrt(0.5 * (1.0 - cos2theta))

            dS_costheta = 0.0
            dS_sintheta = 0.0

            Sig_11_hat = 0.5 * W
            Sig_33_hat = 0.5 * W

            dS_Sig_11_hat = 0.5 * dS_W
            dS_Sig_33_hat = 0.5 * dS_W

        else:
            sqrt_a2_c2 = np.sqrt(a * a + c * c)
            cos2theta = np.abs(2 * a) / (2 * sqrt_a2_c2)
            costheta = np.sqrt(0.5 * (1.0 + cos2theta))
            sintheta = mysign(a) * mysign(c) * np.sqrt(0.5 * (1.0 - cos2theta))

            dS_cos2theta = mysign(a) * (
                0.5 * b / sqrt_a2_c2
                - a
                * (a * b + 2 * c * d)
                / (2 * sqrt_a2_c2 * sqrt_a2_c2 * sqrt_a2_c2)
            )

            dS_costheta = 1 / (4 * costheta) * dS_cos2theta
            if np.abs(sintheta) > threshold_singular:
                # equivalent to: if np.abs(c)>threshold_singular:
                dS_sintheta = -1 / (4 * sintheta) * dS_cos2theta
            else:
                dS_sintheta = d / (2 * a)

            Sig_11_hat = 0.5 * W
            Sig_33_hat = 0.5 * W

            dS_Sig_11_hat = 0.5 * dS_W + mysign(a) * sqrt_a2_c2
            dS_Sig_33_hat = 0.5 * dS_W - mysign(a) * sqrt_a2_c2

    else:

        sqrtT = np.sqrt(T)
        cos2theta = signR * R / sqrtT
        costheta = np.sqrt(0.5 * (1.0 + cos2theta))
        sintheta = signR * mysign(Sig_13) * np.sqrt(0.5 * (1.0 - cos2theta))

        # in sixtrack this line seems to be different different
        # ~ sintheta = -mysign((Sig_11-Sig_33))*np.sqrt(0.5*(1.-cos2theta))

        Sig_11_hat = 0.5 * (W + signR * sqrtT)
        Sig_33_hat = 0.5 * (W - signR * sqrtT)

        dS_cos2theta = signR * (
            dS_R / sqrtT - R / (2 * sqrtT * sqrtT * sqrtT) * dS_T
        )
        dS_costheta = 1 / (4 * costheta) * dS_cos2theta

        if np.abs(sintheta) < threshold_singular and handle_singularities:
            # equivalent to to np.abs(Sig_13)<threshold_singular
            dS_sintheta = (Sig_14 + Sig_23) / R
        else:
            dS_sintheta = -1 / (4 * sintheta) * dS_cos2theta

        dS_Sig_11_hat = 0.5 * (dS_W + signR * 0.5 / sqrtT * dS_T)
        dS_Sig_33_hat = 0.5 * (dS_W - signR * 0.5 / sqrtT * dS_T)

    # This will not be exposed in C
    """
    extra_data = {}
    extra_data['Sig_11'] = Sig_11
    extra_data['Sig_13'] = Sig_13
    extra_data['Sig_33'] = Sig_33
    extra_data['cos2theta'] = cos2theta
    extra_data['T'] = T
    extra_data['R'] = R
    """
    extra_data = -1.0

    # ~ if dS_sintheta>1e10:
    # ~ import pdb; pdb.set_trace()

    return (
        Sig_11_hat,
        Sig_33_hat,
        costheta,
        sintheta,
        dS_Sig_11_hat,
        dS_Sig_33_hat,
        dS_costheta,
        dS_sintheta,
        extra_data,
    )


propagate_Sigma_matrix = np.vectorize(_propagate_Sigma_matrix)

# def propagate_Sigma_matrix(Sigmas_at_0,
#                                       S,
#                                       threshold_singular=1e-16,
#                                       handle_singularities=True):
#
#     Sig_11_hat, Sig_33_hat, costheta, sintheta,\
#         dS_Sig_11_hat, dS_Sig_33_hat, dS_costheta, dS_sintheta,\
#         extra_data_list = np.vectorize(_propagate_Sigma_matrix,
#                                        excluded=['Sigmas_at_0',
#                                                  'threshold_singular',
#                                                  'handle_singularities'])(
#             Sigmas_at_0, S, threshold_singular, handle_singularities)
#
#     extra_data = {}
#     for kk in extra_data_list[0].keys():
#         extra_data[kk] = []
#         for ele in extra_data_list:
#             extra_data[kk].append(ele[kk])
#
#     return Sig_11_hat, Sig_33_hat, costheta, sintheta,\
#         dS_Sig_11_hat, dS_Sig_33_hat, dS_costheta, dS_sintheta,\
#         extra_data
#


def propagate_full_Sigma_matrix_in_drift(
    Sig_11_0,
    Sig_12_0,
    Sig_13_0,
    Sig_14_0,
    Sig_22_0,
    Sig_23_0,
    Sig_24_0,
    Sig_33_0,
    Sig_34_0,
    Sig_44_0,
    S,
):

    # Can be found in matrix form in A. Wolsky,
    # "Beam dynamics in high energy particle accelerators"

    Sig_11 = Sig_11_0 + 2.0 * Sig_12_0 * S + Sig_22_0 * S * S
    Sig_33 = Sig_33_0 + 2.0 * Sig_34_0 * S + Sig_44_0 * S * S
    Sig_13 = Sig_13_0 + (Sig_14_0 + Sig_23_0) * S + Sig_24_0 * S * S

    Sig_12 = Sig_12_0 + Sig_22_0 * S
    Sig_14 = Sig_14_0 + Sig_24_0 * S
    Sig_22 = Sig_22_0 + 0.0 * S
    Sig_23 = Sig_23_0 + Sig_24_0 * S
    Sig_24 = Sig_24_0 + 0.0 * S
    Sig_34 = Sig_34_0 + Sig_44_0 * S
    Sig_44 = Sig_44_0 + 0.0 * S

    return (
        Sig_11,
        Sig_12,
        Sig_13,
        Sig_14,
        Sig_22,
        Sig_23,
        Sig_24,
        Sig_33,
        Sig_34,
        Sig_44,
    )



from xline.mathlibs import MathlibDefault


class QGauss(object):
    @staticmethod
    def calc_cq(q, mathlib=MathlibDefault, EPS=1e-6):
        assert q < 3
        cq = mathlib.sqrt(mathlib.pi)
        if q >= (1 + EPS):
            cq *= mathlib.gamma((3 - q) / (2 * q - 2))
            cq /= mathlib.sqrt((q - 1)) * mathlib.gamma(1 / (q - 1))
        elif q <= (1 - EPS):
            cq *= 2 * mathlib.gamma(1 / (1 - q))
            cq /= (
                (3 - q)
                * mathlib.sqrt(1 - q)
                * mathlib.gamma((3 - q) / (2 - 2 * q))
            )
        return cq

    @staticmethod
    def sqrt_beta(sigma, mathlib=MathlibDefault):
        assert sigma > 0
        return 1 / (mathlib.sqrt(2) * sigma)

    @staticmethod
    def exp_q(x, q, mathlib=MathlibDefault, EPS=1e-6):
        assert q < 3
        if mathlib.abs(1 - q) > EPS:
            u_plus = 1 + x * (1 - q)
            if u_plus < 0:
                u_plus = 0
            return mathlib.pow(u_plus, 1 / (1 - q))
        else:
            return mathlib.exp(x)

    def __init__(self, q=1.0, mathlib=MathlibDefault, cq_eps=1e-6):
        assert q < 3
        self._m = mathlib
        self._q = q
        self._cq = QGauss.calc_cq(q, mathlib=mathlib, EPS=cq_eps)

    @property
    def q(self):
        return self._q

    @q.setter
    def q(self, q_value):
        assert q_value < 3
        self._q = q_value
        self._cq = QGauss.calc_cq(self._q, self._m)

    @property
    def cq(self):
        return self._cq

    def min_support(self, sqrt_beta):
        assert self._q < 3
        assert sqrt_beta > 0
        if self._q >= 1:
            return -1e10
        else:
            return -1 / self._m.sqrt(sqrt_beta * sqrt_beta * (1 - self._q))

    def max_support(self, sqrt_beta):
        return -(self.min_support(sqrt_beta))

    def eval(self, x, sqrt_beta, mu=0.0):
        assert self._m.abs(self._cq) > 0
        assert self._q < 3
        assert sqrt_beta > 0
        factor = sqrt_beta / self._cq
        arg = sqrt_beta * sqrt_beta
        arg *= (x - mu) * (x - mu)
        return factor * QGauss.exp_q(-arg, self._q, self._m)



import numpy as np

from . import elements as xline_elements


def iter_from_madx_sequence(
    sequence,
    classes=xline_elements,
    ignored_madtypes=[],
    exact_drift=False,
    drift_threshold=1e-6,
    install_apertures=False,
):

    if exact_drift:
        myDrift = classes.DriftExact
    else:
        myDrift = classes.Drift
    seq = sequence

    elements = seq.elements
    ele_pos = seq.element_positions()

    old_pp = 0.0
    i_drift = 0
    for pp, ee in sorted(zip(ele_pos,elements),key=lambda x:x[0]):
        skiptilt=False

        if pp > old_pp + drift_threshold:
            yield "drift_%d" % i_drift, myDrift(length=(pp - old_pp))
            old_pp = pp
            i_drift += 1

        eename = ee.name
        mad_etype = ee.base_type.name

        # if ee.length > 0:
        #    raise ValueError(f"Sequence {seq} contains {eename} with length>0")

        if mad_etype in [
            "marker",
            "monitor",
            "hmonitor",
            "vmonitor",
            "collimator",
            "rcollimator",
            "elseparator",
            "instrument",
            "solenoid",
            "drift",
        ]:
            newele = myDrift(length=ee.l)
            old_pp += ee.l

        elif mad_etype in ignored_madtypes:
            pass

        elif mad_etype == "multipole":
            knl = ee.knl if hasattr(ee, "knl") else [0]
            ksl = ee.ksl if hasattr(ee, "ksl") else [0]
            newele = classes.Multipole(
                knl=list(knl),
                ksl=list(ksl),
                hxl=knl[0],
                hyl=ksl[0],
                length=ee.lrad,
            )

        elif mad_etype == "tkicker" or mad_etype == "kicker":
            hkick = [-ee.hkick] if hasattr(ee, "hkick") else []
            vkick = [ee.vkick] if hasattr(ee, "vkick") else []
            newele = classes.Multipole(
                knl=hkick, ksl=vkick, length=ee.lrad, hxl=0, hyl=0
            )

        elif mad_etype == "vkicker":
            newele = classes.Multipole(
                knl=[], ksl=[ee.kick], length=ee.lrad, hxl=0, hyl=0
            )

        elif mad_etype == "hkicker":
            newele = classes.Multipole(
                knl=[-ee.kick], ksl=[], length=ee.lrad, hxl=0, hyl=0
            )
        elif mad_etype == "dipedge":
            newele = classes.DipoleEdge(
                h=ee.h, e1=ee.e1, hgap=ee.hgap, fint=ee.fint
            )

        elif mad_etype == "rfcavity":
            newele = classes.Cavity(
                voltage=ee.volt * 1e6,
                frequency=ee.freq * 1e6,
                lag=ee.lag * 360,
            )

        elif mad_etype == "rfmultipole":
            newele = classes.RFMultipole(
                voltage=ee.volt * 1e6,
                frequency=ee.freq * 1e6,
                lag=ee.lag * 360,
                knl=ee.knl[:],
                ksl=ee.ksl[:],
                pn=[v * 360 for v in ee.pnl],
                ps=[v * 360 for v in ee.psl],
            )

        elif mad_etype == "crabcavity":
            #ee.volt in MV, sequence.beam.pc in GeV
            if abs(ee.tilt-np.pi/2)<1e-9:
                newele = classes.RFMultipole(
                    frequency=ee.freq * 1e6,
                    ksl=[-ee.volt / sequence.beam.pc*1e-3],
                    ps=[ee.lag * 360 + 90],
                )
                skiptilt=True
            else:
                newele = classes.RFMultipole(
                    frequency=ee.freq * 1e6,
                    knl=[ee.volt / sequence.beam.pc*1e-3],
                    pn=[ee.lag * 360 + 90], # TODO: Changed sign to match sixtrack
                                            # To be checked!!!! 

                )


        elif mad_etype == "beambeam":
            if ee.slot_id == 6 or ee.slot_id == 60:
                # BB interaction is 6D
                newele = classes.BeamBeam6D(
                    phi=0.0,
                    alpha=0.0,
                    x_bb_co=0.0,
                    y_bb_co=0.0,
                    charge_slices=[0.0],
                    zeta_slices=[0.0],
                    sigma_11=1.0,
                    sigma_12=0.0,
                    sigma_13=0.0,
                    sigma_14=0.0,
                    sigma_22=1.0,
                    sigma_23=0.0,
                    sigma_24=0.0,
                    sigma_33=0.0,
                    sigma_34=0.0,
                    sigma_44=0.0,
                    x_co=0.0,
                    px_co=0.0,
                    y_co=0.0,
                    py_co=0.0,
                    zeta_co=0.0,
                    delta_co=0.0,
                    d_x=0.0,
                    d_px=0.0,
                    d_y=0.0,
                    d_py=0.0,
                    d_zeta=0.0,
                    d_delta=0.0,
                )
            else:
                # BB interaction is 4D
                newele = classes.BeamBeam4D(
                    charge=0.0,
                    sigma_x=1.0,
                    sigma_y=1.0,
                    beta_r=1.0,
                    x_bb=0.0,
                    y_bb=0.0,
                    d_px=0.0,
                    d_py=0.0,
                )
        elif mad_etype == "placeholder":
            if ee.slot_id == 1:
                newele = classes.SCCoasting()
            elif ee.slot_id == 2:
                newele = classes.SCQGaussProfile()
            elif ee.slot_id == 3:
                newele = classes.SCInterpolatedProfile()
            else:
                newele = myDrift(length=ee.l)
                old_pp += ee.l
        else:
            raise ValueError(f'MAD element "{mad_etype}" not recognized')


        if hasattr(ee,'tilt') and abs(ee.tilt)>0 and not skiptilt:
            tilt=np.rad2deg(ee.tilt)
        else:
            tilt=0

        if abs(tilt)>0:
            yield eename+"_pretilt", classes.SRotation(angle=tilt)

        yield eename, newele

        if abs(tilt)>0:
            yield eename+"_posttilt", classes.SRotation(angle=-tilt)

        if (
            install_apertures
            and hasattr(ee, "aperture")
            and (min(ee.aperture) > 0)
        ):
            if ee.apertype == "rectangle":
                newaperture = classes.LimitRect(
                    min_x=-ee.aperture[0],
                    max_x=ee.aperture[0],
                    min_y=-ee.aperture[1],
                    max_y=ee.aperture[1],
                )
            elif ee.apertype == "ellipse":
                newaperture = classes.LimitEllipse(
                    a=ee.aperture[0], b=ee.aperture[1]
                )
            elif ee.apertype == "circle":
                newaperture = classes.LimitEllipse(
                    a=ee.aperture[0], b=ee.aperture[0]
                )
            elif ee.apertype == "rectellipse":
                newaperture = classes.LimitRectEllipse(
                    max_x=ee.aperture[0],
                    max_y=ee.aperture[1],
                    a=ee.aperture[2],
                    b=ee.aperture[3],
                )
            else:
                raise ValueError("Aperture type not recognized")

            yield eename + "_aperture", newaperture

    if hasattr(seq, "length") and seq.length > old_pp:
        yield "drift_%d" % i_drift, myDrift(length=(seq.length - old_pp))


class MadPoint(object):
    @classmethod
    def from_survey(cls, name, mad):
        return cls(name, mad, use_twiss=False, use_survey=True)

    @classmethod
    def from_twiss(cls, name, mad):
        return cls(name, mad, use_twiss=True, use_survey=False)

    def __init__(self, name, mad, use_twiss=True, use_survey=True):

        self.use_twiss = use_twiss
        self.use_survey = use_survey

        if not (use_survey) and not (use_twiss):
            raise ValueError(
                "use_survey and use_twiss cannot be False at the same time"
            )

        self.name = name
        if use_twiss:
            twiss = mad.table.twiss
            names = twiss.name
        if use_survey:
            survey = mad.table.survey
            names = survey.name

        idx = np.where(names == name)[0][0]

        if use_twiss:
            self.tx = twiss.x[idx]
            self.ty = twiss.y[idx]
            self.tpx = twiss.px[idx]
            self.tpy = twiss.py[idx]
        else:
            self.tx = None
            self.ty = None
            self.tpx = None
            self.tpy = None

        if use_survey:
            self.sx = survey.x[idx]
            self.sy = survey.y[idx]
            self.sz = survey.z[idx]
            self.sp = np.array([self.sx, self.sy, self.sz])
            theta = survey.theta[idx]
            phi = survey.phi[idx]
            psi = survey.psi[idx]
        else:
            self.sx = None
            self.sy = None
            self.sz = None
            self.sp = None
            theta = 0.0
            phi = 0.0
            psi = 0.0

        thetam = np.array(
            [
                [np.cos(theta), 0, np.sin(theta)],
                [0, 1, 0],
                [-np.sin(theta), 0, np.cos(theta)],
            ]
        )
        phim = np.array(
            [
                [1, 0, 0],
                [0, np.cos(phi), np.sin(phi)],
                [0, -np.sin(phi), np.cos(phi)],
            ]
        )
        psim = np.array(
            [
                [np.cos(psi), -np.sin(psi), 0],
                [np.sin(psi), np.cos(psi), 0],
                [0, 0, 1],
            ]
        )
        wm = np.dot(thetam, np.dot(phim, psim))
        self.ex = np.dot(wm, np.array([1, 0, 0]))
        self.ey = np.dot(wm, np.array([0, 1, 0]))
        self.ez = np.dot(wm, np.array([0, 0, 1]))

        self.p = np.array([0.0, 0.0, 0.0])

        if use_twiss:
            self.p += self.ex * self.tx + self.ey * self.ty

        if use_survey:
            self.p += self.sp

    def dist(self, other):
        return np.sqrt(np.sum((self.p - other.p) ** 2))

    def distxy(self, other):
        dd = self.p - other.p
        return np.dot(dd, self.ex), np.dot(dd, self.ey)


def mad_benchmark(mtype, attrs, pc=0.2, x=0, px=0, y=0, py=0, t=0, pt=0):
    import xline
    from cpymad.madx import Madx

    mad = Madx(stdout=False)
    madtype = mad.command[mtype]
    mad.beam(particle="proton", pc=pc)
    madtype.clone("mm", **attrs)
    mad.input("bench: line=(mm)")
    mad.use(sequence="bench")
    mad.track(onepass=True, dump=False)
    mad.start(x=x, px=px, y=y, py=py, t=t, pt=pt)
    mad.run()
    mad.endtrack()
    p_mad = xline.Particles.from_madx_track(mad)
    p_six = p_mad.copy(0)
    line = xline.Line.from_madx_sequence(
        mad.sequence.bench, exact_drift=True
    )
    line.track(p_six)
    p_mad.copy(-1).compare(p_six, rel_tol=0)
    return mad, line, p_mad, p_six



import numpy as np

from . import boost


class ParBoost(object):
    # it is practically a struct
    def __init__(self, phi, alpha):
        self.sphi = np.sin(phi)
        self.cphi = np.cos(phi)
        self.tphi = np.tan(phi)
        self.salpha = np.sin(alpha)
        self.calpha = np.cos(alpha)

    def tobuffer(self):
        buf = [self.sphi, self.cphi, self.tphi, self.salpha, self.calpha]
        return np.array(buf, dtype=np.float64)


class Sigmas(object):
    def __init__(
        self,
        Sig_11_0,
        Sig_12_0,
        Sig_13_0,
        Sig_14_0,
        Sig_22_0,
        Sig_23_0,
        Sig_24_0,
        Sig_33_0,
        Sig_34_0,
        Sig_44_0,
    ):

        self.Sig_11_0 = Sig_11_0
        self.Sig_12_0 = Sig_12_0
        self.Sig_13_0 = Sig_13_0
        self.Sig_14_0 = Sig_14_0
        self.Sig_22_0 = Sig_22_0
        self.Sig_23_0 = Sig_23_0
        self.Sig_24_0 = Sig_24_0
        self.Sig_33_0 = Sig_33_0
        self.Sig_34_0 = Sig_34_0
        self.Sig_44_0 = Sig_44_0

    def tobuffer(self):
        buf = [
            self.Sig_11_0,
            self.Sig_12_0,
            self.Sig_13_0,
            self.Sig_14_0,
            self.Sig_22_0,
            self.Sig_23_0,
            self.Sig_24_0,
            self.Sig_33_0,
            self.Sig_34_0,
            self.Sig_44_0,
        ]
        return np.array(buf, dtype=np.float64)


def boost_sigmas(Sigma_0, cphi):
    Sigma_0_boosted = Sigmas(
        Sigma_0.Sig_11_0,
        Sigma_0.Sig_12_0 / cphi,
        Sigma_0.Sig_13_0,
        Sigma_0.Sig_14_0 / cphi,
        Sigma_0.Sig_22_0 / cphi / cphi,
        Sigma_0.Sig_23_0 / cphi,
        Sigma_0.Sig_24_0 / cphi / cphi,
        Sigma_0.Sig_33_0,
        Sigma_0.Sig_34_0 / cphi,
        Sigma_0.Sig_44_0 / cphi / cphi,
    )
    return Sigma_0_boosted


def int_to_float64arr(val):
    temp = np.zeros(1, (np.float64, {"i64": ("i8", 0)}))
    temp["i64"][0] = val
    return temp


class BB6D_Data(object):
    def __init__(
        self,
        q_part,
        parboost,
        Sigmas_0_star,
        N_slices,
        N_part_per_slice,
        x_slices_star,
        y_slices_star,
        sigma_slices_star,
        min_sigma_diff,
        threshold_singular,
        delta_x,
        delta_y,
        x_CO,
        px_CO,
        y_CO,
        py_CO,
        sigma_CO,
        delta_CO,
        Dx_sub,
        Dpx_sub,
        Dy_sub,
        Dpy_sub,
        Dsigma_sub,
        Ddelta_sub,
        enabled,
    ):

        self.q_part = q_part
        self.parboost = parboost
        self.Sigmas_0_star = Sigmas_0_star
        self.min_sigma_diff = min_sigma_diff
        self.threshold_singular = threshold_singular
        self.N_slices = N_slices
        self.N_part_per_slice = N_part_per_slice
        self.x_slices_star = x_slices_star
        self.y_slices_star = y_slices_star
        self.sigma_slices_star = sigma_slices_star

        self.delta_x = delta_x
        self.delta_y = delta_y
        self.x_CO = x_CO
        self.px_CO = px_CO
        self.y_CO = y_CO
        self.py_CO = py_CO
        self.sigma_CO = sigma_CO
        self.delta_CO = delta_CO
        self.Dx_sub = Dx_sub
        self.Dpx_sub = Dpx_sub
        self.Dy_sub = Dy_sub
        self.Dpy_sub = Dpy_sub
        self.Dsigma_sub = Dsigma_sub
        self.Ddelta_sub = Ddelta_sub

        self.enabled = enabled

    def tobuffer(self):

        buffer_list = []
        # Buffers corresponding to BB6D struct
        buffer_list.append(np.array([self.q_part], dtype=np.float64))
        buffer_list.append(self.parboost.tobuffer())
        buffer_list.append(self.Sigmas_0_star.tobuffer())
        buffer_list.append(np.array([self.min_sigma_diff], dtype=np.float64))
        buffer_list.append(
            np.array([self.threshold_singular], dtype=np.float64)
        )
        buffer_list.append(int_to_float64arr(self.N_slices))

        buffer_list.append(np.array([self.delta_x], dtype=np.float64))
        buffer_list.append(np.array([self.delta_y], dtype=np.float64))
        buffer_list.append(np.array([self.x_CO], dtype=np.float64))
        buffer_list.append(np.array([self.px_CO], dtype=np.float64))
        buffer_list.append(np.array([self.y_CO], dtype=np.float64))
        buffer_list.append(np.array([self.py_CO], dtype=np.float64))
        buffer_list.append(np.array([self.sigma_CO], dtype=np.float64))
        buffer_list.append(np.array([self.delta_CO], dtype=np.float64))
        buffer_list.append(np.array([self.Dx_sub], dtype=np.float64))
        buffer_list.append(np.array([self.Dpx_sub], dtype=np.float64))
        buffer_list.append(np.array([self.Dy_sub], dtype=np.float64))
        buffer_list.append(np.array([self.Dpy_sub], dtype=np.float64))
        buffer_list.append(np.array([self.Dsigma_sub], dtype=np.float64))
        buffer_list.append(np.array([self.Ddelta_sub], dtype=np.float64))
        buffer_list.append(
            int_to_float64arr({True: 1, False: 0}[self.enabled])
        )

        buffer_list.append(int_to_float64arr(3))  # offset to N_part_per_slice
        # offset to x_slices_star
        buffer_list.append(int_to_float64arr(2 + self.N_slices))
        # offset to y_slices_star
        buffer_list.append(int_to_float64arr(1 + 2 * self.N_slices))
        # offset to sigma_slices_star
        buffer_list.append(int_to_float64arr(0 + 3 * self.N_slices))

        # Buffers corresponding to arrays
        buffer_list.append(np.array(self.N_part_per_slice, dtype=np.float64))
        buffer_list.append(np.array(self.x_slices_star, dtype=np.float64))
        buffer_list.append(np.array(self.y_slices_star, dtype=np.float64))
        buffer_list.append(np.array(self.sigma_slices_star, dtype=np.float64))

        buf = np.concatenate(buffer_list)

        return buf


def BB6D_init(
    q_part,
    phi,
    alpha,
    delta_x,
    delta_y,
    N_part_per_slice,
    z_slices,
    Sig_11_0,
    Sig_12_0,
    Sig_13_0,
    Sig_14_0,
    Sig_22_0,
    Sig_23_0,
    Sig_24_0,
    Sig_33_0,
    Sig_34_0,
    Sig_44_0,
    x_CO,
    px_CO,
    y_CO,
    py_CO,
    sigma_CO,
    delta_CO,
    min_sigma_diff,
    threshold_singular,
    Dx_sub,
    Dpx_sub,
    Dy_sub,
    Dpy_sub,
    Dsigma_sub,
    Ddelta_sub,
    enabled,
):

    assert len(N_part_per_slice) == len(z_slices)

    # Prepare data for Lorentz transformation
    parboost = ParBoost(phi=phi, alpha=alpha)

    # Prepare data with strong beam shape
    Sigmas_0 = Sigmas(
        Sig_11_0,
        Sig_12_0,
        Sig_13_0,
        Sig_14_0,
        Sig_22_0,
        Sig_23_0,
        Sig_24_0,
        Sig_33_0,
        Sig_34_0,
        Sig_44_0,
    )

    # Boost strong beam shape
    Sigmas_0_star = boost_sigmas(Sigmas_0, parboost.cphi)

    # Sort according to z, head at the first position in the arrays
    ind_sorted = np.argsort(z_slices)[::-1]
    z_slices = np.take(z_slices, ind_sorted)
    N_part_per_slice = np.take(N_part_per_slice, ind_sorted)

    # By boosting the strong z and all zeros, I get the transverse coordinates
    # of the strong beam in the ref system of the weak
    boost_vect = np.vectorize(boost.boost, excluded="parboost")
    (
        x_slices_star,
        px_slices_star,
        y_slices_star,
        py_slices_star,
        sigma_slices_star,
        delta_slices_star,
    ) = boost_vect(
        x=0 * z_slices,
        px=0 * z_slices,
        y=0 * z_slices,
        py=0 * z_slices,
        sigma=z_slices,
        delta=0 * z_slices,
        parboost=parboost,
    )

    N_slices = len(z_slices)

    bb6d_data = BB6D_Data(
        q_part,
        parboost,
        Sigmas_0_star,
        N_slices,
        N_part_per_slice,
        x_slices_star,
        y_slices_star,
        sigma_slices_star,
        min_sigma_diff,
        threshold_singular,
        delta_x,
        delta_y,
        x_CO,
        px_CO,
        y_CO,
        py_CO,
        sigma_CO,
        delta_CO,
        Dx_sub,
        Dpx_sub,
        Dy_sub,
        Dpy_sub,
        Dsigma_sub,
        Ddelta_sub,
        enabled,
    )

    return bb6d_data



import json
import numpy as np
from dataclasses import dataclass, field
from typing import List, Dict


_function = type(lambda: None)


def _pro_default(default):
    type_default = type(default)
    if type_default is _function:
        mut = default()
        if type(mut) is list:
            return List, field(default_factory=default)
        elif type(mut) is dict:
            return Dict, field(default_factory=default)
    elif type_default in [list, dict]:
        raise ValueError(f"Mutable default {default} not allowed")
    else:
        return type_default, default


class _MetaElement(type):
    def __new__(cls, clsname, bases, dct):
        description = dct.get("_description", [])
        extra = dct.get("_extra", [])
        dct["_base_fields"] = [dd[0] for dd in description]
        dct["_extra_fields"] = [dd[0] for dd in extra]
        dct["_fields"] = dct["_base_fields"] + dct["_extra_fields"]
        ann = {}
        dct["__annotations__"] = ann
        for name, unit, desc, default in description:
            ann[name], dct[name] = _pro_default(default)
        for name, unit, desc, default in extra:
            ann[name], dct[name] = _pro_default(default)
        try:
            doc = [dct["__doc__"], "\nFields:\n"]
        except KeyError:
            doc = ["\nFields:\n"]
        fields = [
            f"    - {name:10} [{unit+']:':5} {desc} "
            for name, unit, desc, default in description
        ]
        fields += [
            f"    - {name:10} [{unit+']:':5} {desc} "
            for name, unit, desc, default in extra
        ]
        doc += fields
        dct["__doc__"] = "\n".join(doc)
        newclass = super(_MetaElement, cls).__new__(cls, clsname, bases, dct)
        return dataclass(newclass)


class Base(metaclass=_MetaElement):

    iscollective = False

    def get_fields(self, keepextra=False):
        if keepextra:
            return self.__class__._fields
        else:
            return self.__class__._base_fields

    def to_dict(self, keepextra=False):
        out = {kk: getattr(self, kk) for kk in self.get_fields(keepextra)}
        out["__class__"] = self.__class__.__name__
        return out

    @classmethod
    def from_dict(cls, dct, keepextra=True):
        self = cls()
        for kk in cls._base_fields:
            setattr(self, kk, dct[kk])
        if keepextra:
            for kk in cls._extra_fields:
                if kk in dct:
                    setattr(self, kk, dct[kk])
        return self

    def copy(self, keepextra=True):
        return self.__class__.from_dict(self.to_dict(keepextra), keepextra)


class Element(Base):
    pass


class JEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        elif np.issubdtype(type(obj), np.integer):
            return int(obj)
        else:
            return json.JSONEncoder.default(self, obj)



from . import propagate_sigma_matrix as psm
from . import boost as boost

# import transverse_efields as tef

from . import gaussian_fields as gf


import numpy as np


from scipy.constants import c as c_light


def BB6D_track(x, px, y, py, sigma, delta, q0, p0, bb6ddata, mathlib):

    # print('Input px',px)

    q_part = bb6ddata.q_part
    parboost = bb6ddata.parboost
    Sigmas_0_star = bb6ddata.Sigmas_0_star
    N_slices = bb6ddata.N_slices
    N_part_per_slice = bb6ddata.N_part_per_slice
    x_slices_star = bb6ddata.x_slices_star
    y_slices_star = bb6ddata.y_slices_star
    sigma_slices_star = bb6ddata.sigma_slices_star
    min_sigma_diff = bb6ddata.min_sigma_diff
    threshold_singular = bb6ddata.threshold_singular

    # Change reference frame
    x_subCO = x - bb6ddata.x_CO - bb6ddata.delta_x
    px_subCO = px - bb6ddata.px_CO
    y_subCO = y - bb6ddata.y_CO - bb6ddata.delta_y
    py_subCO = py - bb6ddata.py_CO
    sigma_subCO = sigma - bb6ddata.sigma_CO
    delta_subCO = delta - bb6ddata.delta_CO

    # Boost coordinates of the weak beam
    x_star, px_star, y_star, py_star, sigma_star, delta_star = boost.boost(
        x_subCO,
        px_subCO,
        y_subCO,
        py_subCO,
        sigma_subCO,
        delta_subCO,
        parboost,
    )
    # ~ x_star, px_star, y_star, py_star, sigma_star, delta_star =
    #      (x, px, y, py, sigma, delta)
    for i_slice in range(N_slices):
        sigma_slice_star = sigma_slices_star[i_slice]
        x_slice_star = x_slices_star[i_slice]
        y_slice_star = y_slices_star[i_slice]

        # Compute force scaling factor
        Ksl = N_part_per_slice[i_slice] * q_part * q0 / (p0 * c_light)

        # Identify the Collision Point (CP)
        S = 0.5 * (sigma_star - sigma_slice_star)

        # Get strong beam shape at the CP
        (
            Sig_11_hat_star,
            Sig_33_hat_star,
            costheta,
            sintheta,
            dS_Sig_11_hat_star,
            dS_Sig_33_hat_star,
            dS_costheta,
            dS_sintheta,
            extra_data,
        ) = psm.propagate_Sigma_matrix(
            Sigmas_0_star, S, threshold_singular=threshold_singular
        )

        # Evaluate transverse coordinates of the weake baem w.r.t.
        # the strong beam centroid
        x_bar_star = x_star + px_star * S - x_slice_star
        y_bar_star = y_star + py_star * S - y_slice_star

        # Move to the uncoupled reference frame
        x_bar_hat_star = x_bar_star * costheta + y_bar_star * sintheta
        y_bar_hat_star = -x_bar_star * sintheta + y_bar_star * costheta

        # Compute derivatives of the transformation
        dS_x_bar_hat_star = x_bar_star * dS_costheta + y_bar_star * dS_sintheta
        dS_y_bar_hat_star = (
            -x_bar_star * dS_sintheta + y_bar_star * dS_costheta
        )

        # Compute normalized field
        # Ex, Ey, Gx, Gy = tef.get_Ex_Ey_Gx_Gy_gauss(
        #                     x=x_bar_hat_star, y=y_bar_hat_star,
        #                     sigma_x=np.sqrt(Sig_11_hat_star),
        #                     sigma_y=np.sqrt(Sig_33_hat_star),
        #                     min_sigma_diff = min_sigma_diff)

        Ex, Ey, Gx, Gy = gf.get_Ex_Ey_Gx_Gy_gauss(
            x=x_bar_hat_star,
            y=y_bar_hat_star,
            sigma_x=np.sqrt(Sig_11_hat_star),
            sigma_y=np.sqrt(Sig_33_hat_star),
            min_sigma_diff=min_sigma_diff,
            skip_Gs=False,
            mathlib=mathlib,
        )

        # Compute kicks
        Fx_hat_star = Ksl * Ex
        Fy_hat_star = Ksl * Ey
        Gx_hat_star = Ksl * Gx
        Gy_hat_star = Ksl * Gy

        # Move kisks to coupled reference frame
        Fx_star = Fx_hat_star * costheta - Fy_hat_star * sintheta
        Fy_star = Fx_hat_star * sintheta + Fy_hat_star * costheta

        # Compute longitudinal kick
        Fz_star = 0.5 * (
            Fx_hat_star * dS_x_bar_hat_star
            + Fy_hat_star * dS_y_bar_hat_star
            + Gx_hat_star * dS_Sig_11_hat_star
            + Gy_hat_star * dS_Sig_33_hat_star
        )

        # Apply the kicks (Hirata's synchro-beam)
        delta_star = (
            delta_star
            + Fz_star
            + 0.5
            * (
                Fx_star * (px_star + 0.5 * Fx_star)
                + Fy_star * (py_star + 0.5 * Fy_star)
            )
        )
        x_star = x_star - S * Fx_star
        px_star = px_star + Fx_star
        y_star = y_star - S * Fy_star
        py_star = py_star + Fy_star

    # Inverse boost on the coordinates of the weak beam
    x_ret, px_ret, y_ret, py_ret, sigma_ret, delta_ret = boost.inv_boost(
        x_star, px_star, y_star, py_star, sigma_star, delta_star, parboost
    )

    # Go back to original reference frame and remove dipolar effect
    x_out = x_ret + bb6ddata.x_CO + bb6ddata.delta_x - bb6ddata.Dx_sub
    px_out = px_ret + bb6ddata.px_CO - bb6ddata.Dpx_sub
    y_out = y_ret + bb6ddata.y_CO + bb6ddata.delta_y - bb6ddata.Dy_sub
    py_out = py_ret + bb6ddata.py_CO - bb6ddata.Dpy_sub
    sigma_out = sigma_ret + bb6ddata.sigma_CO - bb6ddata.Dsigma_sub
    delta_out = delta_ret + bb6ddata.delta_CO - bb6ddata.Ddelta_sub

    # print(x_ret, px_ret, y_ret, py_ret, sigma_ret, delta_ret)
    return x_out, px_out, y_out, py_out, sigma_out, delta_out



import numpy as np
from scipy.special import erfinv


def constant_charge_slicing_gaussian(N_part_tot, sigmaz, N_slices):

    if N_slices > 1:
        # working with intensity 1. and rescling at the end
        Qi = (np.arange(N_slices) / float(N_slices))[1:]

        z_cuts = np.sqrt(2) * sigmaz * erfinv(2 * Qi - 1.0)

        # ~ import pdb; pdb.set_trace()

        z_centroids = []
        first_centroid = (
            -sigmaz
            / np.sqrt(2 * np.pi)
            * np.exp(-z_cuts[0] ** 2 / (2 * sigmaz * sigmaz))
            * float(N_slices)
        )
        z_centroids.append(first_centroid)
        for ii in range(N_slices - 2):
            this_centroid = (
                -sigmaz
                / np.sqrt(2 * np.pi)
                * (
                    np.exp(-z_cuts[ii + 1] ** 2 / (2 * sigmaz * sigmaz))
                    - np.exp(-z_cuts[ii] ** 2 / (2 * sigmaz * sigmaz))
                )
                * float(N_slices)
            )
            # the multiplication times n slices comes from the fact that
            # we have to divide by the slice charge, i.e. 1./N
            z_centroids.append(this_centroid)

        last_centroid = (
            sigmaz
            / np.sqrt(2 * np.pi)
            * np.exp(-z_cuts[-1] ** 2 / (2 * sigmaz * sigmaz))
            * float(N_slices)
        )
        z_centroids.append(last_centroid)

        z_centroids = np.array(z_centroids)

        N_part_per_slice = z_centroids * 0.0 + N_part_tot / float(N_slices)
    elif N_slices == 1:
        z_centroids = np.array([0.0])
        z_cuts = []
        N_part_per_slice = np.array([N_part_tot])

    else:
        raise ValueError("Invalid number of slices")

    return z_centroids, z_cuts, N_part_per_slice



__version__ = "0.0.9"

from . import base_classes
from . import elements
from .elements import *

from .line import Line

from .particles import Particles

from .be_beamfields import BB6Ddata
from .loader_mad import MadPoint, mad_benchmark




from scipy.special import wofz
from scipy.special import gamma as tgamma


class MathlibDefault(object):

    from numpy import sqrt, exp, sin, cos, abs, pi, tan, interp, linspace
    from numpy import power as pow

    @classmethod
    def wfun(cls, z_re, z_im):
        w = wofz(z_re + 1j * z_im)
        return w.real, w.imag

    @classmethod
    def gamma(cls, arg):
        assert arg > 0.0
        return tgamma(arg)



import json
import numpy as np

from .base_classes import Element, JEncoder
from . import elements
from .particles import Particles

from .loader_sixtrack import _expand_struct
from .loader_mad import iter_from_madx_sequence
from .closed_orbit import linearize_around_closed_orbit
from .closed_orbit import healy_symplectify
from .linear_normal_form import _linear_normal_form



_thick_element_types = (elements.Drift, elements.DriftExact)

def _is_thick(element):
    return  ((hasattr(element, "isthick") and element.isthick) or
             (isinstance(element, _thick_element_types)))


# missing access to particles._m:
deg2rad = np.pi / 180.


class Line(Element):
    _description = [
        ("elements", "", "List of elements", ()),
        ("element_names", "", "List of element names", ()),
    ]
    _extra = []

    def __init__(self, elements=(), element_names=None):
        self.elements = elements
        if element_names is None:
            element_names = [ f"e{ii}" for ii in range(len(elements))]

        self.element_names = element_names

        assert len(self.elements) == len(self.element_names)

    def __len__(self):
        assert len(self.elements) == len(self.element_names)
        return len(self.elements)


    def to_dict(self, keepextra=True):
        out = {}
        out["elements"] = [el.to_dict(keepextra) for el in self.elements]
        out["element_names"] = self.element_names[:]
        return out

    @classmethod
    def from_dict(cls, dct, keepextra=True):
        self = cls(elements=[], element_names=[])
        for el in dct["elements"]:
            eltype = getattr(elements, el["__class__"])
            newel = eltype.from_dict(el)
            self.elements.append(newel)
        self.element_names = dct["element_names"]
        return self

    def to_json(self, filename,  keepextra=True):
        with open(filename, 'w') as fid:
            json.dump(self.to_dict(keepextra=keepextra), fid, cls=JEncoder)

    @classmethod
    def from_json(cls, filename,  keepextra=True):
        with open(filename, 'r') as fid:
            return cls.from_dict(json.load(fid), keepextra=keepextra)

    def append_line(self, line):
        # Append the elements
        if type(line) is Line:
            # got a xline line
            self.elements += line.elements
        else:
            # got a different type of line (e.g. pybplep)
            for ee in line.elements:
                type_name = ee.__class__.__name__
                newele = getattr(elements, type_name)(**ee._asdict())
                self.elements.append(newele)

        # Append the names
        self.element_names += line.element_names

        assert len(self.elements) == len(self.element_names)

        return self

    def track(self, p):
        ret = None
        for el in self.elements:
            ret = el.track(p)
            if ret is not None:
                break
        return ret

    def track_elem_by_elem(self, p, start=True, end=False):
        out = []
        if start:
            out.append(p.copy())
        for el in self.elements:
            ret = el.track(p)
            if ret is not None:
                break
            out.append(p.copy())
        if end:
            out.append(p.copy())
        return out

    def insert_element(self, idx, element, name):
        self.elements.insert(idx, element)
        self.element_names.insert(idx, name)
        # assert len(self.elements) == len(self.element_names)
        return self

    def append_element(self, element, name):
        self.elements.append(element)
        self.element_names.append(name)
        # assert len(self.elements) == len(self.element_names)
        return self

    def get_length(self):
        thick_element_types = (elements.Drift, elements.DriftExact)

        ll = 0
        for ee in self.elements:
            if _is_thick(ee):
                ll += ee.length

        return ll

    def get_s_elements(self, mode="upstream"):

        assert mode in ["upstream", "downstream"]
        s_prev = 0
        s = []
        for ee in self.elements:
            if mode == "upstream":
                s.append(s_prev)
            if _is_thick(ee):
                s_prev += ee.length
            if mode == "downstream":
                s.append(s_prev)
        return s

    def remove_inactive_multipoles(self, inplace=False):
        newline = Line(elements=[], element_names=[])

        for ee, nn in zip(self.elements, self.element_names):
            if isinstance(ee, (elements.Multipole)):
                aux = [ee.hxl, ee.hyl] + list(ee.knl) + list(ee.ksl)
                if np.sum(np.abs(np.array(aux))) == 0.0:
                    continue
            newline.append_element(ee, nn)

        if inplace:
            self.elements.clear()
            self.element_names.clear()
            self.append_line(newline)
            return self
        else:
            return newline

    def remove_zero_length_drifts(self, inplace=False):
        newline = Line(elements=[], element_names=[])

        for ee, nn in zip(self.elements, self.element_names):
            if isinstance(ee, (elements.Drift, elements.DriftExact)):
                if ee.length == 0.0:
                    continue
            newline.append_element(ee, nn)

        if inplace:
            self.elements.clear()
            self.element_names.clear()
            self.append_line(newline)
            return self
        else:
            return newline

    def merge_consecutive_drifts(self, inplace=False):
        newline = Line(elements=[], element_names=[])

        for ee, nn in zip(self.elements, self.element_names):
            if len(newline.elements) == 0:
                newline.append_element(ee, nn)
                continue

            if isinstance(ee, (elements.Drift, elements.DriftExact)):
                prev_ee = newline.elements[-1]
                prev_nn = newline.element_names[-1]
                if isinstance(prev_ee, (elements.Drift, elements.DriftExact)):
                    prev_ee.length += ee.length
                    prev_nn += ('_' + nn)
                    newline.element_names[-1] = prev_nn
                else:
                    newline.append_element(ee, nn)
            else:
                newline.append_element(ee, nn)

        if inplace:
            self.elements.clear()
            self.element_names.clear()
            self.append_line(newline)
            return self
        else:
            return newline

    def merge_consecutive_multipoles(self, inplace=False):
        newline = Line(elements=[], element_names=[])

        for ee, nn in zip(self.elements, self.element_names):
            if len(newline.elements) == 0:
                newline.append_element(ee, nn)
                continue

            if isinstance(ee, elements.Multipole):
                prev_ee = newline.elements[-1]
                prev_nn = newline.element_names[-1]
                if isinstance(prev_ee, elements.Multipole) and prev_ee.hxl==ee.hxl and prev_ee.hyl==ee.hyl:
                    oo=max(len(prev_ee.knl),len(prev_ee.ksl),len(ee.knl),len(ee.ksl))
                    knl=np.zeros(oo,dtype=float)
                    ksl=np.zeros(oo,dtype=float)
                    for ii,kk in enumerate(prev_ee.knl):
                        knl[ii]+=kk
                    for ii,kk in enumerate(ee.knl):
                        knl[ii]+=kk
                    for ii,kk in enumerate(prev_ee.ksl):
                        ksl[ii]+=kk
                    for ii,kk in enumerate(ee.ksl):
                        ksl[ii]+=kk
                    prev_ee.knl=knl
                    prev_ee.ksl=ksl
                    prev_nn += ('_' + nn)
                    newline.element_names[-1] = prev_nn
                else:
                    newline.append_element(ee, nn)
            else:
                newline.append_element(ee, nn)

        if inplace:
            self.elements.clear()
            self.element_names.clear()
            self.append_line(newline)
            return self
        else:
            return newline
    def get_elements_of_type(self, types):
        if not hasattr(types, "__iter__"):
            type_list = [types]
        else:
            type_list = types

        names = []
        elements = []
        for ee, nn in zip(self.elements, self.element_names):
            for tt in type_list:
                if isinstance(ee, tt):
                    names.append(nn)
                    elements.append(ee)

        return elements, names

    def get_element_ids_of_type(self, types, start_idx_offset=0):
        assert start_idx_offset >= 0
        if not hasattr(types, "__iter__"):
            type_list = [types]
        else:
            type_list = types
        elem_idx = []
        for idx, elem in enumerate(self.elements):
            for tt in type_list:
                if isinstance(elem, tt):
                    elem_idx.append(idx+start_idx_offset)
                    break
        return elem_idx

    def linear_normal_form(self, M):
        return _linear_normal_form(M)

    def find_closed_orbit_and_linear_OTM(
        self, p0c, guess=None, d=1.e-7, tol=1.e-10, max_iterations=20, longitudinal_coordinate='zeta'
    ):
        if guess is None:
            guess = [0., 0., 0., 0., 0., 0.]
        
        assert len(guess) == 6

        closed_orbit = np.array(guess).copy()
    
        canonical_conjugate_momentum = {'tau' : 'ptau', 'zeta' : 'delta', 'sigma' : 'psigma'}
    
        if longitudinal_coordinate not in ['tau', 'zeta', 'sigma']:
            raise Exception('Longitudinal variable not recognized in search of closed orbit')
    
        longitudinal_momentum = canonical_conjugate_momentum[longitudinal_coordinate]
    
        for i in range(max_iterations):
            new_closed_orbit, M = linearize_around_closed_orbit(
                self, closed_orbit, p0c, d, longitudinal_coordinate, longitudinal_momentum
            )

            error = np.linalg.norm( new_closed_orbit - closed_orbit )
    
            closed_orbit = new_closed_orbit
            if error < tol:
                print('Converged with approximate distance: {}'.format(error))
                _, M = linearize_around_closed_orbit(
                    self, closed_orbit, p0c, d, longitudinal_coordinate, longitudinal_momentum
                )
                return closed_orbit, healy_symplectify(M)
    
            print ('Closed orbit search iteration: {}'.format(i))
    
        print('WARNING!: Search did not converge, approximate distance: {}'.format(error))
        return closed_orbit, healy_symplectify(M)

    def find_closed_orbit(
            self, p0c, guess=[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            method="Nelder-Mead", **kwargs
            ):
        def _one_turn_map(coord):
            pcl = Particles(p0c=p0c, **kwargs)
            pcl.x = coord[0]
            pcl.px = coord[1]
            pcl.y = coord[2]
            pcl.py = coord[3]
            pcl.zeta = coord[4]
            pcl.delta = coord[5]

            self.track(pcl)
            coord_out = np.array(
                [pcl.x, pcl.px, pcl.y, pcl.py, pcl.zeta, pcl.delta]
            )

            return coord_out

        def _CO_error(coord):
            return np.sum((_one_turn_map(coord) - coord) ** 2)

        if method == "get_guess":
            res = type("", (), {})()
            res.x = guess
        else:
            import scipy.optimize as so

            res = so.minimize(
                _CO_error, np.array(guess), tol=1e-20, method=method
            )

        pcl = Particles(p0c=p0c, **kwargs)

        pcl.x = res.x[0]
        pcl.px = res.x[1]
        pcl.y = res.x[2]
        pcl.py = res.x[3]
        pcl.zeta = res.x[4]
        pcl.delta = res.x[5]

        return pcl

    def enable_beambeam(self):

        for ee in self.elements:
            if isinstance(ee, (elements.BeamBeam4D, elements.BeamBeam6D)):
                ee.enabled = True

    def disable_beambeam(self):

        for ee in self.elements:
            if isinstance(ee, (elements.BeamBeam4D, elements.BeamBeam6D)):
                ee.enabled = False

    def beambeam_store_closed_orbit_and_dipolar_kicks(
        self,
        particle_on_CO,
        separation_given_wrt_closed_orbit_4D=True,
        separation_given_wrt_closed_orbit_6D=True,
    ):

        self.disable_beambeam()
        closed_orbit = self.track_elem_by_elem(particle_on_CO)

        self.enable_beambeam()

        for ie, ee in enumerate(self.elements):
            # to transfer to beambeam.py

            if ee.__class__.__name__ == "BeamBeam4D":
                if separation_given_wrt_closed_orbit_4D:
                    ee.x_bb += closed_orbit[ie].x
                    ee.y_bb += closed_orbit[ie].y

                # Evaluate dipolar kick
                ptemp = closed_orbit[ie].copy()
                ptempin = ptemp.copy()

                ee.track(ptemp)

                ee.d_px = ptemp.px - ptempin.px
                ee.d_py = ptemp.py - ptempin.py

            elif ee.__class__.__name__ == "BeamBeam6D":
                if not separation_given_wrt_closed_orbit_6D:
                    raise ValueError("Not implemented!")

                # Store closed orbit
                ee.x_co = closed_orbit[ie].x
                ee.px_co = closed_orbit[ie].px
                ee.y_co = closed_orbit[ie].y
                ee.py_co = closed_orbit[ie].py
                ee.zeta_co = closed_orbit[ie].zeta
                ee.delta_co = closed_orbit[ie].delta

                # Evaluate 6d kick on closed orbit
                ptemp = closed_orbit[ie].copy()
                ptempin = ptemp.copy()

                ee.track(ptemp)

                ee.d_x = ptemp.x - ptempin.x
                ee.d_px = ptemp.px - ptempin.px
                ee.d_y = ptemp.y - ptempin.y
                ee.d_py = ptemp.py - ptempin.py
                ee.d_zeta = ptemp.zeta - ptempin.zeta
                ee.d_delta = ptemp.delta - ptempin.delta

    @classmethod
    def from_sixinput(cls, sixinput, classes=elements):
        other_info = {}

        line_data, rest, iconv = _expand_struct(sixinput, convert=classes)

        ele_names = [dd[0] for dd in line_data]
        elements = [dd[2] for dd in line_data]

        line = cls(elements=elements, element_names=ele_names)

        other_info["rest"] = rest
        other_info["iconv"] = iconv

        line.other_info = other_info

        return line

    @classmethod
    def from_madx_sequence(
        cls,
        sequence,
        classes=elements,
        ignored_madtypes=[],
        exact_drift=False,
        drift_threshold=1e-6,
        install_apertures=False,
        apply_madx_errors=False,
    ):

        line = cls(elements=[], element_names=[])

        for el_name, el in iter_from_madx_sequence(
            sequence,
            classes=classes,
            ignored_madtypes=ignored_madtypes,
            exact_drift=exact_drift,
            drift_threshold=drift_threshold,
            install_apertures=install_apertures,
        ):
            line.append_element(el, el_name)

        if apply_madx_errors:
            line._apply_madx_errors(sequence)

        return line

    # error handling (alignment, multipole orders, ...):

    def find_element_ids(self, element_name):
        """Find element_name in this Line instance's
        self.elements_name list. Assumes the names are unique.

        Return index before and after the element, taking into account
        attached _aperture instances (LimitRect, LimitEllipse, ...)
        which would follow the element occurrence in the list.

        Raises IndexError if element_name not found in this Line.
        """
        # will raise error if element not present:
        idx_el = self.element_names.index(element_name)
        try:
            # if aperture marker is present
            idx_after_el = self.element_names.index(element_name + "_aperture") + 1
        except ValueError:
            # if aperture marker is not present
            idx_after_el = idx_el + 1
        return idx_el, idx_after_el

    def _add_offset_error_to(self, element_name, dx=0, dy=0):
        idx_el, idx_after_el = self.find_element_ids(element_name)
        xyshift = elements.XYShift(dx=dx, dy=dy)
        inv_xyshift = elements.XYShift(dx=-dx, dy=-dy)
        self.insert_element(idx_el, xyshift, element_name + "_offset_in")
        self.insert_element(
            idx_after_el + 1, inv_xyshift, element_name + "_offset_out"
        )

    def _add_aperture_offset_error_to(self, element_name, arex=0, arey=0):
        idx_el, idx_after_el = self.find_element_ids(element_name)
        idx_el_aper = idx_after_el - 1
        if not self.element_names[idx_el_aper] == element_name + "_aperture":
            # it is allowed to provide arex/arey without providing an aperture
            print('Info: Element', element_name, ': arex/y provided without aperture -> arex/y ignored')
            return
        xyshift = elements.XYShift(dx=arex, dy=arey)
        inv_xyshift = elements.XYShift(dx=-arex, dy=-arey)
        self.insert_element(idx_el_aper, xyshift, element_name + "_aperture_offset_in")
        self.insert_element(
            idx_after_el + 1, inv_xyshift, element_name + "_aperture_offset_out"
        )

    def _add_tilt_error_to(self, element_name, angle):
        '''Alignment error of transverse rotation around s-axis.
        The element corresponding to the given `element_name`
        gets wrapped by SRotation elements with rotation angle
        `angle`.

        In the case of a thin dipole component, the corresponding
        curvature terms in the Multipole (hxl and hyl) are rotated
        by `angle` as well.
        '''
        idx_el, idx_after_el = self.find_element_ids(element_name)
        element = self.elements[self.element_names.index(element_name)]
        if isinstance(element, elements.Multipole) and (
                element.hxl or element.hyl):
            dpsi = angle * deg2rad

            hxl0 = element.hxl
            hyl0 = element.hyl

            hxl1 = hxl0 * np.cos(dpsi) - hyl0 * np.sin(dpsi)
            hyl1 = hxl0 * np.sin(dpsi) + hyl0 * np.cos(dpsi)

            element.hxl = hxl1
            element.hyl = hyl1
        srot = elements.SRotation(angle=angle)
        inv_srot = elements.SRotation(angle=-angle)
        self.insert_element(idx_el, srot, element_name + "_tilt_in")
        self.insert_element(idx_after_el + 1, inv_srot, element_name + "_tilt_out")

    def _add_multipole_error_to(self, element_name, knl=[], ksl=[]):
        # will raise error if element not present:
        assert element_name in self.element_names
        element = self.elements[self.element_names.index(element_name)]
        # normal components
        knl = np.trim_zeros(knl, trim="b")
        if len(element.knl) < len(knl):
            element.knl += [0] * (len(knl) - len(element.knl))
        for i, component in enumerate(knl):
            element.knl[i] += component
        # skew components
        ksl = np.trim_zeros(ksl, trim="b")
        if len(element.ksl) < len(ksl):
            element.ksl += [0] * (len(ksl) - len(element.ksl))
        for i, component in enumerate(ksl):
            element.ksl[i] += component

    def _apply_madx_errors(self, madx_sequence):
        """Applies errors from MAD-X sequence to existing
        elements in this Line instance.

        Return names of MAD-X elements with existing align_errors
        or field_errors which were not found in the elements of
        this Line instance (and thus not treated).

        Example via cpymad:
            madx = cpymad.madx.Madx()

            # (...set up lattice and errors in cpymad...)

            seq = madx.sequence.some_lattice
            xline_line = xline.Line.from_madx_sequence(
                                    seq,
                                    apply_madx_errors=True
                              )
        """
        elements_not_found = []
        for element, element_name in zip(
                madx_sequence.expanded_elements,
                madx_sequence.expanded_element_names()
        ):
            if element_name not in self.element_names:
                if element.align_errors or element.field_errors:
                    elements_not_found.append(element_name)
                    continue

            if element.align_errors:
                # add offset
                dx = element.align_errors.dx
                dy = element.align_errors.dy
                if dx or dy:
                    self._add_offset_error_to(element_name, dx, dy)

                # add tilt
                dpsi = element.align_errors.dpsi
                if dpsi:
                    self._add_tilt_error_to(element_name, angle=dpsi / deg2rad)

                # add aperture-only offset
                arex = element.align_errors.arex
                arey = element.align_errors.arey
                if arex or arey:
                    self._add_aperture_offset_error_to(element_name, arex, arey)

                # check for errors which cannot be treated yet:
                #for error_type in dir(element.align_errors):
                 #   if not error_type[0] == '_' and \
                  #          error_type not in ['dx', 'dy', 'dpsi', 'arex',
                   #                            'arey', 'count', 'index']:
                        #print(
                        #    f'Warning: MAD-X error type "{error_type}"'
                        #    " not implemented yet."
                        #)

            if element.field_errors:
                # add multipole error
                if any(element.field_errors.dkn) or \
                            any(element.field_errors.dks):
                    knl = element.field_errors.dkn
                    ksl = element.field_errors.dks
                    on=np.where(knl)[0]
                    os=np.where(ksl)[0]
                    on = on[-1] if len(on)>0 else 0
                    os = os[-1] if len(os)>0 else 0
                    oo = max(os,on)+1
                    knl = knl[:oo]  # delete trailing zeros
                    ksl = ksl[:oo]  # to keep order low
                    self._add_multipole_error_to(element_name, knl, ksl)

        return elements_not_found


elements.Line = Line




import numpy as np
from .particles import Particles


def get_init_particles_for_linear_map(
        closed_orbit, p0c, d, longitudinal_coordinate,
        longitudinal_momentum, **kwargs):
    part = Particles(p0c=p0c, **kwargs)
#    part = Particles()
#    part.p0c = p0c
    part.x  = closed_orbit[0] + np.array([0.0, 1.0 * d, 0.0, 0.0, 0.0, 0.0, 0.0])
    part.px = closed_orbit[1] + np.array([0.0, 0.0, 1.0 * d, 0.0, 0.0, 0.0, 0.0])
    part.y  = closed_orbit[2] + np.array([0.0, 0.0, 0.0, 1.0 * d, 0.0, 0.0, 0.0])
    part.py = closed_orbit[3] + np.array([0.0, 0.0, 0.0, 0.0, 1.0 * d, 0.0, 0.0])

    longi   = closed_orbit[4] + np.array([0.0, 0.0, 0.0, 0.0, 0.0, 1.0 * d, 0.0])
    plongi  = closed_orbit[5] + np.array([0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0 * d])

    setattr(part, longitudinal_coordinate, longi)
    setattr(part, longitudinal_momentum, plongi)

    return part


def part_to_array(part, longitudinal_coordinate, longitudinal_momentum):
    X_array = np.empty([6, 7])
    X_array[0, :] = part.x
    X_array[1, :] = part.px
    X_array[2, :] = part.y
    X_array[3, :] = part.py
    X_array[4, :] = getattr(part, longitudinal_coordinate)
    X_array[5, :] = getattr(part, longitudinal_momentum)

    return X_array


def linearize_around_closed_orbit(
        line, closed_orbit, p0c, d, longitudinal_coordinate,
        longitudinal_momentum, **kwargs):

    part = get_init_particles_for_linear_map(
        closed_orbit, p0c, d, longitudinal_coordinate,
        longitudinal_momentum, **kwargs)
    X_init = part_to_array(
        part, longitudinal_coordinate, longitudinal_momentum)

    line.track(part)
    X_fin = part_to_array(
        part, longitudinal_coordinate, longitudinal_momentum)

    m = X_fin[:, 0] - X_init[:, 0]
    M = np.empty([6, 6])
    for j in range(6):
        M[:, j] = (X_fin[:, j + 1] - X_fin[:, 0]) / d

    X_CO = X_init[:, 0] + np.matmul(
        np.linalg.inv(np.identity(6) - M), m.T)

    return X_CO, M


def healy_symplectify(M):
    # https://accelconf.web.cern.ch/e06/PAPERS/WEPCH152.PDF
    print("Symplectifying linear One-Turn-Map...")

    print("Before symplectifying: det(M) = {}".format(np.linalg.det(M)))
    I = np.identity(6)

    S = np.array(
        [
            [0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
            [-1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, -1.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0, 1.0],
            [0.0, 0.0, 0.0, 0.0, -1.0, 0.0],
        ]
    )

    V = np.matmul(S, np.matmul(I - M, np.linalg.inv(I + M)))
    W = (V + V.T) / 2
    if np.linalg.det(I - np.matmul(S, W)) != 0:
        M_new = np.matmul(I + np.matmul(S, W),
                          np.linalg.inv(I - np.matmul(S, W)))
    else:
        print("WARNING: det(I - SW) = 0!")
        V_else = np.matmul(S, np.matmul(I + M, np.linalg.inv(I - M)))
        W_else = (V_else + V_else.T) / 2
        M_new = -np.matmul(
            I + np.matmul(S, W_else), np.linalg(I - np.matmul(S, W_else))
        )

    print("After symplectifying: det(M) = {}".format(np.linalg.det(M_new)))
    return M_new



import numpy as np
from scipy.constants import e as qe
from scipy.constants import c as clight

from xline.base_classes import Element
from .gaussian_fields import get_Ex_Ey_Gx_Gy_gauss
from . import BB6D
from . import BB6Ddata


class BeamBeam4D(Element):
    """Interaction with a transverse-Gaussian strong beam (4D modelling)."""

    _description = [
        (
            "charge",
            "e",
            "Charge of the interacting distribution (strong beam)",
            0,
        ),
        ("sigma_x", "m", "Horizontal size of the strong beam (r.m.s.)", 1.0),
        ("sigma_y", "m", "Vertical size of the strong beam (r.m.s.)", 1.0),
        ("beta_r", "", "Relativistic beta of the stron beam", 1.0),
        (
            "x_bb",
            "m",
            "H. position of the strong beam w.r.t. the reference trajectory",
            0,
        ),
        (
            "y_bb",
            "m",
            "V. position of the strong beam w.r.t. the reference trajectory",
            0,
        ),
        ("d_px", "", "H. kick subtracted after the interaction.", 0),
        ("d_py", "", "V. kick subtracted after the interaction.", 0),
    ]

    _extra = [
        ("min_sigma_diff", "m", "Threshold to detect round beam", 1e-28),
        ("enabled", "", "Switch for closed orbit search", True),
    ]

    def track(self, p):
        if self.enabled:
            charge = p.charge_ratio * p.q0
            x = p.x - self.x_bb
            px = p.px
            y = p.y - self.y_bb
            py = p.py

            chi = p.chi

            beta = p.beta0 / p.rvv
            p0c = p.p0c * qe

            Ex, Ey = get_Ex_Ey_Gx_Gy_gauss(
                x,
                y,
                self.sigma_x,
                self.sigma_y,
                min_sigma_diff=1e-10,
                skip_Gs=True,
                mathlib=p._m,
            )

            fact_kick = (
                chi
                * self.charge
                * qe
                * charge
                * qe
                * (1.0 + beta * self.beta_r)
                / (p0c * (beta + self.beta_r))
            )

            px += fact_kick * Ex - self.d_px
            py += fact_kick * Ey - self.d_py

            p.px = px
            p.py = py


class BeamBeam6D(Element):
    """Interaction with a transverse-Gaussian strong beam (6D modelling).

    http://cds.cern.ch/record/2306400
    """

    _description = [
        (
            "phi",
            "rad",
            "Crossing angle (>0 weak beam increases"
            "x in the direction motion)",
            0,
        ),
        ("alpha", "rad", "Crossing plane tilt angle (>0 x tends to y)", 0),
        (
            "x_bb_co",
            "m",
            "H. position of the strong beam w.r.t. the closed orbit",
            0,
        ),
        (
            "y_bb_co",
            "m",
            "V. position of the strong beam w.r.t. the closed orbit",
            0,
        ),
        (
            "charge_slices",
            "qe",
            "Charge of the interacting slices (strong beam)",
            (0.0),
        ),
        (
            "zeta_slices",
            "m",
            "Longitudinal position of the interacting"
            "slices (>0 head of the strong).",
            (0.0),
        ),
        (
            "sigma_11",
            "m^2",
            "Sigma_11 element of the sigma matrix of the strong beam",
            1.0,
        ),
        (
            "sigma_12",
            "m",
            "Sigma_12 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_13",
            "m^2",
            "Sigma_13 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_14",
            "m",
            "Sigma_14 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_22",
            "",
            "Sigma_22 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_23",
            "m",
            "Sigma_23 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_24",
            "",
            "Sigma_24 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_33",
            "m^2",
            "Sigma_33 element of the sigma matrix of the strong beam",
            1.0,
        ),
        (
            "sigma_34",
            "m",
            "Sigma_34 element of the sigma matrix of the strong beam",
            0,
        ),
        (
            "sigma_44",
            "",
            "Sigma_44 element of the sigma matrix of the strong beam",
            0,
        ),
        ("x_co", "m", "x coordinate the closed orbit (weak beam).", 0),
        ("px_co", "", "px coordinate the closed orbit (weak beam).", 0),
        ("y_co", "m", "y coordinate the closed orbit (weaek beam).", 0),
        ("py_co", "", "py coordinate the closed orbit (weaek beam).", 0),
        ("zeta_co", "m", "zeta coordinate the closed orbit (weaek beam).", 0),
        ("delta_co", "", "delta coordinate the closed orbit (weaek beam).", 0),
        ("d_x", "m", "Quantity subtracted from x after the interaction.", 0),
        ("d_px", "", "Quantity subtracted from px after the interaction.", 0),
        ("d_y", "m", "Quantity subtracted from y after the interaction.", 0),
        ("d_py", "", "Quantity subtracted from py after the interaction.", 0),
        (
            "d_zeta",
            "m",
            "Quantity subtracted from sigma after the interaction.",
            0,
        ),
        (
            "d_delta",
            "",
            "Quantity subtracted from delta after the interaction.",
            0,
        ),
    ]
    _extra = [
        ("min_sigma_diff", "m", "Threshold to detect round beam", 1e-28),
        (
            "threshold_singular",
            "",
            "Threshold to detect small denominator",
            1e-28,
        ),
        ("enabled", "", "Switch for closed orbit search", True),
    ]

    def track(self, p):
        if self.enabled:
            bb6data = BB6Ddata.BB6D_init(
                qe,
                self.phi,
                self.alpha,
                self.x_bb_co,
                self.y_bb_co,
                np.atleast_1d(self.charge_slices),
                np.atleast_1d(self.zeta_slices),
                self.sigma_11,
                self.sigma_12,
                self.sigma_13,
                self.sigma_14,
                self.sigma_22,
                self.sigma_23,
                self.sigma_24,
                self.sigma_33,
                self.sigma_34,
                self.sigma_44,
                self.x_co,
                self.px_co,
                self.y_co,
                self.py_co,
                self.zeta_co,
                self.delta_co,
                self.min_sigma_diff,
                self.threshold_singular,
                self.d_x,
                self.d_px,
                self.d_y,
                self.d_py,
                self.d_zeta,
                self.d_delta,
                self.enabled,
            )
            (
                x_ret,
                px_ret,
                y_ret,
                py_ret,
                zeta_ret,
                delta_ret,
            ) = BB6D.BB6D_track(
                p.x,
                p.px,
                p.y,
                p.py,
                p.zeta,
                p.delta,
                p.q0 * qe,
                p.p0c / clight * qe,
                bb6data,
                mathlib=p._m,
            )
            self._last_bb6data = bb6data
            p.x = x_ret
            p.px = px_ret
            p.y = y_ret
            p.py = py_ret
            p.zeta = zeta_ret
            p.delta = delta_ret



import numpy as np
import xline
from xline import MadPoint

_sigma_names = [11, 12, 13, 14, 22, 23, 24, 33, 34, 44]
_beta_names = ["betx", "bety"]


def norm(v):
    return np.sqrt(np.sum(v ** 2))


def get_points_twissdata_for_elements(
    ele_names, mad, seq_name, use_survey=True, use_twiss=True
):

    mad.use(sequence=seq_name)
    mad.twiss()

    if use_survey:
        mad.survey()

    bb_xyz_points = []
    bb_twissdata = {
        kk: []
        for kk in _sigma_names
        + _beta_names
        + "dispersion_x dispersion_y x y".split()
    }
    for eename in ele_names:
        bb_xyz_points.append(
            MadPoint(
                eename + ":1", mad, use_twiss=use_twiss, use_survey=use_survey
            )
        )

        i_twiss = np.where(mad.table.twiss.name == (eename + ":1"))[0][0]

        for sn in _sigma_names:
            bb_twissdata[sn].append(
                getattr(mad.table.twiss, "sig%d" % sn)[i_twiss]
            )

        for kk in ["betx", "bety"]:
            bb_twissdata[kk].append(mad.table.twiss[kk][i_twiss])
        gamma = mad.table.twiss.summary.gamma
        beta = np.sqrt(1.0 - 1.0 / (gamma * gamma))
        for pp in ["x", "y"]:
            bb_twissdata["dispersion_" + pp].append(
                mad.table.twiss["d" + pp][i_twiss] * beta
            )
            bb_twissdata[pp].append(mad.table.twiss[pp][i_twiss])
        # , 'dx', 'dy']:

    return bb_xyz_points, bb_twissdata


def get_elements(seq, ele_type=None, slot_id=None):

    elements = []
    element_names = []
    for ee in seq.elements:

        if ele_type is not None:
            if ee.base_type.name != ele_type:
                continue

        if slot_id is not None:
            if ee.slot_id != slot_id:
                continue

        elements.append(ee)
        element_names.append(ee.name)

    return elements, element_names


def get_points_twissdata_for_element_type(
    mad, seq_name, ele_type=None, slot_id=None, use_survey=True, use_twiss=True
):

    elements, element_names = get_elements(
        seq=mad.sequence[seq_name], ele_type=ele_type, slot_id=slot_id
    )

    points, twissdata = get_points_twissdata_for_elements(
        element_names,
        mad,
        seq_name,
        use_survey=use_survey,
        use_twiss=use_twiss,
    )

    return elements, element_names, points, twissdata


###############################
# beam beam related functions #
###############################
def find_alpha_and_phi(dpx, dpy):

    phi = np.sqrt(dpx ** 2 + dpy ** 2) / 2.0
    if phi < 1e-20:
        alpha = 0.0
    elif np.abs(dpx) >= np.abs(dpy):
        alpha = np.arctan(dpy / dpx)
        if dpx < 0:
            phi = -phi
    else:
        alpha = np.sign(dpy) * (np.pi / 2 - np.abs(np.arctan(dpx / dpy)))
        if dpy < 0:
            phi = -phi

    return alpha, phi


def get_bb_names_madpoints_sigmas(
    mad, seq_name, use_survey=True, use_twiss=True
):
    (
        _,
        element_names,
        points,
        twissdata,
    ) = get_points_twissdata_for_element_type(
        mad,
        seq_name,
        ele_type="beambeam",
        slot_id=None,
        use_survey=use_survey,
        use_twiss=use_twiss,
    )
    sigmas = {kk: twissdata[kk] for kk in _sigma_names}
    return element_names, points, sigmas


def shift_strong_beam_based_on_close_ip(
    points_weak, points_strong, IPs_survey_weak, IPs_survey_strong
):

    for i_bb, _ in enumerate(points_weak):

        pbw = points_weak[i_bb]
        pbs = points_strong[i_bb]

        # Find closest IP
        d_ip = 1e6
        use_ip = 0
        for ip in IPs_survey_weak.keys():
            dd = norm(pbw.p - IPs_survey_weak[ip].p)
            if dd < d_ip:
                use_ip = ip
                d_ip = dd

        # Shift Bs
        shift_ws = IPs_survey_strong[use_ip].p - IPs_survey_weak[use_ip].p
        pbs.p -= shift_ws


def find_bb_separations(points_weak, points_strong, names=None):

    if names is None:
        names = ["bb_%d" % ii for ii in range(len(points_weak))]

    sep_x = []
    sep_y = []
    for i_bb, name_bb in enumerate(names):

        pbw = points_weak[i_bb]
        pbs = points_strong[i_bb]

        # Find vws
        vbb_ws = points_strong[i_bb].p - points_weak[i_bb].p

        # Check that the two reference system are parallel
        try:
            assert norm(pbw.ex - pbs.ex) < 1e-10  # 1e-4 is a reasonable limit
            assert norm(pbw.ey - pbs.ey) < 1e-10  # 1e-4 is a reasonable limit
            assert norm(pbw.ez - pbs.ez) < 1e-10  # 1e-4 is a reasonable limit
        except AssertionError:
            print(name_bb, "Reference systems are not parallel")
            if (
                np.sqrt(
                    norm(pbw.ex - pbs.ex) ** 2
                    + norm(pbw.ey - pbs.ey) ** 2
                    + norm(pbw.ez - pbs.ez) ** 2
                )
                < 5e-3
            ):
                print("Smaller that 5e-3, tolerated.")
            else:
                raise ValueError("Too large! Stopping.")

        # Check that there is no longitudinal separation
        try:
            assert np.abs(np.dot(vbb_ws, pbw.ez)) < 1e-4
        except AssertionError:
            print(name_bb, "The beams are longitudinally shifted")

        # Find separations
        sep_x.append(np.dot(vbb_ws, pbw.ex))
        sep_y.append(np.dot(vbb_ws, pbw.ey))

    return sep_x, sep_y


def setup_beam_beam_in_line(
    line,
    bb_names,
    bb_sigmas_strong,
    bb_points_weak,
    bb_points_strong,
    beta_r_strong,
    bunch_intensity_strong,
    n_slices_6D,
    bb_coupling,
):

    sep_x, sep_y = find_bb_separations(
        points_weak=bb_points_weak,
        points_strong=bb_points_strong,
        names=bb_names,
    )

    i_bb = 0
    assert bb_coupling is False  # Not implemented
    for ee, eename in zip(line.elements, line.element_names):
        if isinstance(ee, xline.elements.BeamBeam4D):
            assert eename == bb_names[i_bb]

            ee.charge = bunch_intensity_strong
            ee.sigma_x = np.sqrt(bb_sigmas_strong[11][i_bb])
            ee.sigma_y = np.sqrt(bb_sigmas_strong[33][i_bb])
            ee.beta_r = beta_r_strong
            ee.x_bb = sep_x[i_bb]
            ee.y_bb = sep_y[i_bb]

            i_bb += 1
        if isinstance(ee, xline.elements.BeamBeam6D):
            assert eename == bb_names[i_bb]

            dpx = bb_points_weak[i_bb].tpx - bb_points_strong[i_bb].tpx
            dpy = bb_points_weak[i_bb].tpy - bb_points_strong[i_bb].tpy

            alpha, phi = find_alpha_and_phi(dpx, dpy)

            ee.phi = phi
            ee.alpha = alpha
            ee.x_bb_co = sep_x[i_bb]
            ee.y_bb_co = sep_y[i_bb]
            ee.charge_slices = [bunch_intensity_strong / n_slices_6D]
            ee.zeta_slices = [0.0]
            ee.sigma_11 = bb_sigmas_strong[11][i_bb]
            ee.sigma_12 = bb_sigmas_strong[12][i_bb]
            ee.sigma_13 = bb_sigmas_strong[13][i_bb]
            ee.sigma_14 = bb_sigmas_strong[14][i_bb]
            ee.sigma_22 = bb_sigmas_strong[22][i_bb]
            ee.sigma_23 = bb_sigmas_strong[23][i_bb]
            ee.sigma_24 = bb_sigmas_strong[24][i_bb]
            ee.sigma_33 = bb_sigmas_strong[33][i_bb]
            ee.sigma_34 = bb_sigmas_strong[34][i_bb]
            ee.sigma_44 = bb_sigmas_strong[44][i_bb]

            if not (bb_coupling):
                ee.sigma_13 = 0.0
                ee.sigma_14 = 0.0
                ee.sigma_23 = 0.0
                ee.sigma_24 = 0.0

            i_bb += 1


##################################
# space charge related functions #
##################################
sc_mode_to_slotid = {"Coasting": "1", "Bunched": "2", "Interpolated": "3"}


def determine_sc_locations(line, n_SCkicks, length_fuzzy):
    s_elements = np.array(line.get_s_elements())
    length_target = s_elements[-1] / float(n_SCkicks)
    s_targets = np.arange(0, s_elements[-1], length_target)
    sc_locations = []
    for s in s_targets:
        idx_closest = (np.abs(s_elements - s)).argmin()
        s_closest = s_elements[idx_closest]
        if abs(s - s_closest) < length_fuzzy / 2.0:
            sc_locations.append(s_closest)
        else:
            sc_locations.append(s)
    sc_lengths = np.diff(sc_locations).tolist() + [
        s_elements[-1] - sc_locations[-1]
    ]
    return sc_locations, sc_lengths


def install_sc_placeholders(mad, seq_name, name, s, mode="Bunched"):
    sid = sc_mode_to_slotid[mode]
    mad.input(
        f"""
            seqedit, sequence={seq_name};"""
    )
    for name_, s_ in zip(np.atleast_1d(name), np.atleast_1d(s)):
        mad.input(
            f"""
            {name_} : placeholder, l=0., slot_id={sid};
            install, element={name_}, at={s_:.10e};"""
        )
    mad.input(
        f"""
            flatten;
            endedit;
            use, sequence={seq_name};"""
    )


def get_spacecharge_names_twdata(mad, seq_name, mode):
    _, mad_sc_names, _, twdata = get_points_twissdata_for_element_type(
        mad,
        seq_name,
        ele_type="placeholder",
        slot_id=int(sc_mode_to_slotid[mode]),
        use_survey=False,
        use_twiss=True,
    )
    return mad_sc_names, twdata


def _setup_spacecharge_in_line(
    sc_elements,
    sc_lengths,
    sc_twdata,
    betagamma,
    number_of_particles,
    delta_rms,
    neps_x,
    neps_y,
):

    for ii, ss in enumerate(sc_elements):
        ss.number_of_particles = number_of_particles
        ss.sigma_x = np.sqrt(
            sc_twdata["betx"][ii] * neps_x / betagamma
            + (sc_twdata["dispersion_x"][ii] * delta_rms) ** 2
        )
        ss.sigma_y = np.sqrt(
            sc_twdata["bety"][ii] * neps_y / betagamma
            + (sc_twdata["dispersion_y"][ii] * delta_rms) ** 2
        )
        ss.length = sc_lengths[ii]
        ss.x_co = sc_twdata["x"][ii]
        ss.y_co = sc_twdata["y"][ii]
        ss.enabled = True


def setup_spacecharge_bunched_in_line(
    sc_elements,
    sc_lengths,
    sc_twdata,
    betagamma,
    number_of_particles,
    delta_rms,
    neps_x,
    neps_y,
    bunchlength_rms,
):

    for ii, ss in enumerate(sc_elements):
        ss.bunchlength_rms = bunchlength_rms
    _setup_spacecharge_in_line(
        sc_elements,
        sc_lengths,
        sc_twdata,
        betagamma,
        number_of_particles,
        delta_rms,
        neps_x,
        neps_y,
    )


def setup_spacecharge_coasting_in_line(
    sc_elements,
    sc_lengths,
    sc_twdata,
    betagamma,
    number_of_particles,
    delta_rms,
    neps_x,
    neps_y,
    circumference,
):

    for ii, ss in enumerate(sc_elements):
        ss.circumference = circumference
    _setup_spacecharge_in_line(
        sc_elements,
        sc_lengths,
        sc_twdata,
        betagamma,
        number_of_particles,
        delta_rms,
        neps_x,
        neps_y,
    )


def setup_spacecharge_interpolated_in_line(
    sc_elements,
    sc_lengths,
    sc_twdata,
    betagamma,
    number_of_particles,
    delta_rms,
    neps_x,
    neps_y,
    line_density_profile,
    dz,
    z0,
    method=0,
):
    assert method == 0 or method == 1
    for ii, ss in enumerate(sc_elements):
        ss.line_density_profile = line_density_profile
        ss.dz = dz
        ss.z0 = z0
        ss.method = method
    _setup_spacecharge_in_line(
        sc_elements,
        sc_lengths,
        sc_twdata,
        betagamma,
        number_of_particles,
        delta_rms,
        neps_x,
        neps_y,
    )


def check_spacecharge_consistency(
    sc_elements, sc_names, sc_lengths, mad_sc_names
):
    assert len(sc_elements) == len(mad_sc_names)
    assert len(sc_lengths) == len(mad_sc_names)
    for ii, (ss, nn) in enumerate(zip(sc_elements, sc_names)):
        assert nn == mad_sc_names[ii]



from scipy.constants import epsilon_0
import numpy as np


def _get_transv_field_gauss_round(sigma, Delta_x, Delta_y, x, y, mathlib):
    exp = mathlib.exp
    sqrt = mathlib.sqrt
    pi = mathlib.pi

    r2 = (x - Delta_x) * (x - Delta_x) + (y - Delta_y) * (y - Delta_y)
    if r2 < 1e-20:
        temp = sqrt(r2) / (2.0 * pi * epsilon_0 * sigma)  # linearised
    else:
        temp = (1 - exp(-0.5 * r2 / (sigma * sigma))) / (
            2.0 * pi * epsilon_0 * r2
        )

    Ex = temp * (x - Delta_x)
    Ey = temp * (y - Delta_y)

    return Ex, Ey


get_transv_field_gauss_round = np.vectorize(
    _get_transv_field_gauss_round, excluded=["mathlib"]
)


def _get_transv_field_gauss_ellip(
    sigmax, sigmay, Delta_x, Delta_y, x, y, mathlib
):

    abs = mathlib.abs
    exp = mathlib.exp
    wfun = mathlib.wfun
    sqrt = mathlib.sqrt
    pi = mathlib.pi
    sqrt_pi = sqrt(pi)

    # I always go to the first quadrant and then apply the signs a posteriori
    # numerically more stable
    # (see http://inspirehep.net/record/316705/files/slac-pub-5582.pdf)

    abx = abs(x - Delta_x)
    aby = abs(y - Delta_y)

    if sigmax > sigmay:

        S = sqrt(2.0 * (sigmax * sigmax - sigmay * sigmay))
        factBE = 1.0 / (2.0 * epsilon_0 * sqrt_pi * S)

        etaBE_re = sigmay / sigmax * abx
        etaBE_im = sigmax / sigmay * aby

        zetaBE_re = abx
        zetaBE_im = aby

        w_zetaBE_re, w_zetaBE_im = wfun(zetaBE_re / S, zetaBE_im / S)
        w_etaBE_re, w_etaBE_im = wfun(etaBE_re / S, etaBE_im / S)

        expBE = exp(
            -abx * abx / (2 * sigmax * sigmax)
            - aby * aby / (2 * sigmay * sigmay)
        )

        Ex = factBE * (w_zetaBE_im - w_etaBE_im * expBE)
        Ey = factBE * (w_zetaBE_re - w_etaBE_re * expBE)

    elif sigmay > sigmax:

        S = sqrt(2.0 * (sigmay * sigmay - sigmax * sigmax))
        factBE = 1.0 / (2.0 * epsilon_0 * sqrt_pi * S)

        etaBE_re = sigmax / sigmay * aby
        etaBE_im = sigmay / sigmax * abx

        zetaBE_re = aby
        zetaBE_im = abx

        w_zetaBE_re, w_zetaBE_im = wfun(zetaBE_re / S, zetaBE_im / S)
        w_etaBE_re, w_etaBE_im = wfun(etaBE_re / S, etaBE_im / S)

        expBE = exp(
            -aby * aby / (2 * sigmay * sigmay)
            - abx * abx / (2 * sigmax * sigmax)
        )

        Ey = factBE * (w_zetaBE_im - w_etaBE_im * expBE)
        Ex = factBE * (w_zetaBE_re - w_etaBE_re * expBE)
    else:
        Ex = 1.0 / 0.0
        Ey = 1.0 / 0.0

    if (x - Delta_x) < 0:
        Ex = -Ex
    if (y - Delta_y) < 0:
        Ey = -Ey

    return Ex, Ey


get_transv_field_gauss_ellip = np.vectorize(
    _get_transv_field_gauss_ellip, excluded=["mathlib"]
)


def _get_Ex_Ey_Gx_Gy_gauss(
    x, y, sigma_x, sigma_y, min_sigma_diff, skip_Gs, mathlib
):

    abs = mathlib.abs
    pi = mathlib.pi
    exp = mathlib.exp

    if abs(sigma_x - sigma_y) < min_sigma_diff:

        sigma = 0.5 * (sigma_x + sigma_y)
        Delta_x = 0.0
        Delta_y = 0.0

        Ex, Ey = get_transv_field_gauss_round(
            sigma, Delta_x, Delta_y, x, y, mathlib
        )

        if not skip_Gs:
            if abs(x) + abs(y) < min_sigma_diff:
                Gx = 0.0
                Gy = 0.0
            else:
                Gx = (
                    1
                    / (2.0 * (x * x + y * y))
                    * (
                        y * Ey
                        - x * Ex
                        + 1.0
                        / (2 * pi * epsilon_0 * sigma * sigma)
                        * x
                        * x
                        * exp(-(x * x + y * y) / (2.0 * sigma * sigma))
                    )
                )
                Gy = (
                    1.0
                    / (2 * (x * x + y * y))
                    * (
                        x * Ex
                        - y * Ey
                        + 1.0
                        / (2 * pi * epsilon_0 * sigma * sigma)
                        * y
                        * y
                        * exp(-(x * x + y * y) / (2.0 * sigma * sigma))
                    )
                )
    else:

        sigma_x = sigma_x
        sigma_y = sigma_y
        Delta_x = 0.0
        Delta_y = 0.0

        Ex, Ey = get_transv_field_gauss_ellip(
            sigma_x, sigma_y, Delta_x, Delta_y, x, y, mathlib
        )

        Sig_11 = sigma_x * sigma_x
        Sig_33 = sigma_y * sigma_y
        if not skip_Gs:
            Gx = (
                -1.0
                / (2 * (Sig_11 - Sig_33))
                * (
                    x * Ex
                    + y * Ey
                    + 1.0
                    / (2 * pi * epsilon_0)
                    * (
                        sigma_y
                        / sigma_x
                        * exp(-x * x / (2 * Sig_11) - y * y / (2 * Sig_33))
                        - 1.0
                    )
                )
            )
            Gy = (
                1.0
                / (2 * (Sig_11 - Sig_33))
                * (
                    x * Ex
                    + y * Ey
                    + 1.0
                    / (2 * pi * epsilon_0)
                    * (
                        sigma_x
                        / sigma_y
                        * exp(-x * x / (2 * Sig_11) - y * y / (2 * Sig_33))
                        - 1.0
                    )
                )
            )

    if skip_Gs:
        return Ex, Ey
    else:
        return Ex, Ey, Gx, Gy


get_Ex_Ey_Gx_Gy_gauss = np.vectorize(
    _get_Ex_Ey_Gx_Gy_gauss, excluded=["mathlib"]
)



import setuptools
import sys

requirements = {"install": ["numpy"]}

if sys.version_info < (3, 7):
    requirements["install"].append("dataclasses")

version = open("xline/__init__.py").readline().split('"')[1]

setuptools.setup(
    name="xline",
    version=version,
    description="6D Tracking Code",
    author="Riccardo De Maria, Giovanni Iadarola",
    author_email="riccardo.de.maria@cern.ch",
    url="https://github.com/rdemaria/xline",
    packages=["xline", "xline.be_beamfields"],
    package_dir={"xline": "xline"},
    install_requires=requirements["install"],
)



import numpy as np

# I program as close as possible to C...


def boost(x, px, y, py, sigma, delta, parboost):

    sphi = parboost.sphi
    cphi = parboost.cphi
    tphi = parboost.tphi
    salpha = parboost.salpha
    calpha = parboost.calpha

    h = (
        delta
        + 1.0
        - np.sqrt((1.0 + delta) * (1.0 + delta) - px * px - py * py)
    )

    px_st = px / cphi - h * calpha * tphi / cphi
    py_st = py / cphi - h * salpha * tphi / cphi
    delta_st = (
        delta - px * calpha * tphi - py * salpha * tphi + h * tphi * tphi
    )

    pz_st = np.sqrt(
        (1.0 + delta_st) * (1.0 + delta_st) - px_st * px_st - py_st * py_st
    )
    hx_st = px_st / pz_st
    hy_st = py_st / pz_st
    hsigma_st = 1.0 - (delta_st + 1) / pz_st

    L11 = 1.0 + hx_st * calpha * sphi
    L12 = hx_st * salpha * sphi
    L13 = calpha * tphi

    L21 = hy_st * calpha * sphi
    L22 = 1.0 + hy_st * salpha * sphi
    L23 = salpha * tphi

    L31 = hsigma_st * calpha * sphi
    L32 = hsigma_st * salpha * sphi
    L33 = 1.0 / cphi

    x_st = L11 * x + L12 * y + L13 * sigma
    y_st = L21 * x + L22 * y + L23 * sigma
    sigma_st = L31 * x + L32 * y + L33 * sigma

    return x_st, px_st, y_st, py_st, sigma_st, delta_st


def inv_boost(x_st, px_st, y_st, py_st, sigma_st, delta_st, parboost):

    sphi = parboost.sphi
    cphi = parboost.cphi
    tphi = parboost.tphi
    salpha = parboost.salpha
    calpha = parboost.calpha

    pz_st = np.sqrt(
        (1.0 + delta_st) * (1.0 + delta_st) - px_st * px_st - py_st * py_st
    )
    hx_st = px_st / pz_st
    hy_st = py_st / pz_st
    hsigma_st = 1.0 - (delta_st + 1) / pz_st

    Det_L = (
        1.0 / cphi
        + (hx_st * calpha + hy_st * salpha - hsigma_st * sphi) * tphi
    )

    Linv_11 = (
        1.0 / cphi + salpha * tphi * (hy_st - hsigma_st * salpha * sphi)
    ) / Det_L
    Linv_12 = (salpha * tphi * (hsigma_st * calpha * sphi - hx_st)) / Det_L
    Linv_13 = (
        -tphi
        * (
            calpha
            - hx_st * salpha * salpha * sphi
            + hy_st * calpha * salpha * sphi
        )
        / Det_L
    )

    Linv_21 = (calpha * tphi * (-hy_st + hsigma_st * salpha * sphi)) / Det_L
    Linv_22 = (
        1.0 / cphi + calpha * tphi * (hx_st - hsigma_st * calpha * sphi)
    ) / Det_L
    Linv_23 = (
        -tphi
        * (
            salpha
            - hy_st * calpha * calpha * sphi
            + hx_st * calpha * salpha * sphi
        )
        / Det_L
    )

    Linv_31 = -hsigma_st * calpha * sphi / Det_L
    Linv_32 = -hsigma_st * salpha * sphi / Det_L
    Linv_33 = (1.0 + hx_st * calpha * sphi + hy_st * salpha * sphi) / Det_L

    x_i = Linv_11 * x_st + Linv_12 * y_st + Linv_13 * sigma_st
    y_i = Linv_21 * x_st + Linv_22 * y_st + Linv_23 * sigma_st
    sigma_i = Linv_31 * x_st + Linv_32 * y_st + Linv_33 * sigma_st

    h = (delta_st + 1.0 - pz_st) * cphi * cphi

    px_i = px_st * cphi + h * calpha * tphi
    py_i = py_st * cphi + h * salpha * tphi

    delta_i = (
        delta_st
        + px_i * calpha * tphi
        + py_i * salpha * tphi
        - h * tphi * tphi
    )

    return x_i, px_i, y_i, py_i, sigma_i, delta_i
